import os
import numpy as np
from datetime import datetime, timedelta, timezone
from ..time_format import forday, dayfor

def bool_str(value):
    if value:
        vstr = 'T'
    else:
        vstr = 'F'
    return vstr

def blkdat(m):
    ##generate blkdat.input based on model object m
    nmlstr =  f"ECWMF forcing; flx-s14w; LWcorr; precip+2mm; SSSrlx; FCT2 tsadvc.; 0-tracer.\n"
    nmlstr += f"Sigma0; GDEM3 init; KPP mixed layer; SeaWiFS mon KPAR; nested in ATLd0.08 2.6;\n"
    nmlstr += f"S-Z(15-11): dp00/f/x/i=3m/1.125/12m/1m; ds=1m/1.125/4m; src_2.2.12;\n"
    nmlstr += f"12345678901234567890123456789012345678901234567890123456789012345678901234567890\n"
    nmlstr += f" {m.V[0]}{m.V[2]}       'iversn' = hycom version number x10\n"
    nmlstr += f" {m.E}      'iexpt ' = experiment number x10\n"
    nmlstr += f"{m.idm:4d}      'idm   ' = longitudinal array size\n"
    nmlstr += f"{m.jdm:4d}      'jdm   ' = latitudinal  array size\n"
    nmlstr += f"{m.itest:4d}      'itest ' = grid point where detailed diagnostics are desired\n"
    nmlstr += f"{m.jtest:4d}      'jtest ' = grid point where detailed diagnostics are desired\n"
    nmlstr += f"{m.kdm:4d}      'kdm   ' = number of layers\n"
    nmlstr += f"{m.nhybrd:4d}      'nhybrd' = number of hybrid levels (0=all isopycnal)\n"
    nmlstr += f"{m.nsigma:4d}      'nsigma' = number of sigma  levels (nhybrd-nsigma z-levels)\n"
    nmlstr += f"{str(m.dp00):^8s}  'dp00  ' = deep    z-level spacing minimum thickness (m)\n"
    nmlstr += f"{str(m.dp00x):^8s}  'dp00x ' = deep    z-level spacing maximum thickness (m)\n"
    nmlstr += f"{str(m.dp00f):^8s}  'dp00f ' = deep    z-level spacing stretching factor (1.0=const.space)\n"
    nmlstr += f"{str(m.ds00):^8s}  'ds00  ' = shallow z-level spacing minimum thickness (m)\n"
    nmlstr += f"{str(m.ds00x):^8s}  'ds00x ' = shallow z-level spacing maximum thickness (m)\n"
    nmlstr += f"{str(m.ds00f):^8s}  'ds00f ' = shallow z-level spacing stretching factor (1.0=const.space)\n"
    nmlstr += f"{str(m.dp00i):^8s}  'dp00i ' = deep iso-pycnal spacing minimum thickness (m)\n"
    nmlstr += f"{str(m.isotop):^8s}  'isotop' = shallowest depth for isopycnal layers     (m, <0 from file)\n"
    nmlstr += f"{str(m.saln0):^8s}  'saln0 ' = initial salinity value (psu), only used for iniflg<2\n"
    nmlstr += f"{str(m.locsig):^8s}  'locsig' = locally-referenced pot. density for stability (0=F,1=T)\n"
    nmlstr += f"{str(m.kapref):^8s}  'kapref' = thermobaric ref. state (-1=input,0=none,1,2,3=constant)\n"
    nmlstr += f"{str(m.thflag):^8s}  'thflag' = reference pressure flag (0=Sigma-0, 2=Sigma-2)\n"
    nmlstr += f"{str(m.thbase):^8s}  'thbase' = reference density (sigma units)\n"
    nmlstr += f"{str(m.vsigma):^8s}  'vsigma' = spacially varying isopycnal target densities (0=F,1=T)\n"
    for k in range(m.kdm):
        nmlstr += f"{str(m.sigma[k]):^8s}  'sigma ' = layer  {k+1}  density (sigma units)\n"
    nmlstr += f"{str(m.iniflg):^8s}  'iniflg' = initial state flag (0=levl, 1=zonl, 2=clim)\n"
    nmlstr += f"{str(m.jerlv0):^8s}  'jerlv0' = initial jerlov water type (1 to 5; 0 to use KPAR)\n"
    nmlstr += f"{str(m.yrflag):^8s}  'yrflag' = days in year flag   (0=360,  1=366,  2=366J1, 3=actual)\n"
    nmlstr += f"{str(m.sshflg):^8s}  'sshflg' = diagnostic SSH flag (0=SSH,1=SSH&stericSSH)\n"
    nmlstr += f"{str(m.dsurfq):^8s}  'dsurfq' = number of days between model diagnostics at the surface\n"
    nmlstr += f"{str(m.diagfq):^8s}  'diagfq' = number of days between model diagnostics\n"
    nmlstr += f"{str(m.proffq):^8s}  'proffq' = number of days between model diagnostics at selected locs\n"
    nmlstr += f"{str(m.tilefq):^8s}  'tilefq' = number of days between model diagnostics on selected tiles\n"
    nmlstr += f"{str(m.meanfq):^8s}  'meanfq' = number of days between model diagnostics (time averaged)\n"
    nmlstr += f"{str(m.rstrfq):^8s}  'rstrfq' = number of days between model restart output\n"
    nmlstr += f"{str(m.bnstfq):^8s}  'bnstfq' = number of days between baro nesting archive input\n"
    nmlstr += f"{str(m.nestfq):^8s}  'nestfq' = number of days between 3-d  nesting archive input\n"
    nmlstr += f"{str(m.cplifq):^8s}  'cplifq' = number of days (or time steps) between sea ice coupling\n"
    nmlstr += f"{str(m.baclin):^8s}  'baclin' = baroclinic time step (seconds), int. divisor of 86400\n"
    nmlstr += f"{str(m.batrop):^8s}  'batrop' = barotropic time step (seconds), int. div. of baclin/2\n"
    nmlstr += f"{str(m.incflg):^8s}  'incflg' = incremental update flag (0=no, 1=yes, 2=full-velocity)\n"
    nmlstr += f"{str(m.incstp):^8s}  'incstp' = no. timesteps for full update (1=direct insertion)\n"
    nmlstr += f"{str(m.incupf):^8s}  'incupf' = number of days of incremental updating input\n"
    nmlstr += f"{str(m.ra2fac):^8s}  'ra2fac' = weight for Robert-Asselin time filter\n"
    nmlstr += f"{str(m.wbaro ):^8s}  'wbaro ' = barotropic time smoothing weight\n"
    nmlstr += f"{str(m.btrlfr):^8s}  'btrlfr' = leapfrog barotropic time step (0=F,1=T)\n"
    nmlstr += f"{str(m.btrmas):^8s}  'btrmas' = barotropic is mass conserving (0=F,1=T)\n"
    nmlstr += f"{str(m.hybraf):^8s}  'hybraf' = HYBGEN: Robert-Asselin flag   (0=F,1=T)\n"
    nmlstr += f"{str(m.hybrlx):^8s}  'hybrlx' = HYBGEN: inverse relaxation coefficient (time steps)\n"
    nmlstr += f"{str(m.hybiso):^8s}  'hybiso' = HYBGEN: Use PCM if layer is within hybiso of target density\n"
    nmlstr += f"{str(m.hybmap):^8s}  'hybmap' = hybrid   remapper  flag (0=PCM, 1=PLM,  2=PPM)\n"
    nmlstr += f"{str(m.hybflg):^8s}  'hybflg' = hybrid   generator flag (0=T&S, 1=th&S, 2=th&T)\n"
    nmlstr += f"{str(m.advflg):^8s}  'advflg' = thermal  advection flag (0=T&S, 1=th&S, 2=th&T)\n"
    nmlstr += f"{str(m.advtyp):^8s}  'advtyp' = scalar   advection type (0=PCM,1=MPDATA,2=FCT2,4=FCT4)\n"
    nmlstr += f"{str(m.momtyp):^8s}  'momtyp' = momentum advection type (2=2nd order, 4=4th order)\n"
    nmlstr += f"{str(m.slip  ):^8s}  'slip  ' = +1 for free-slip, -1 for non-slip boundary conditions\n"
    nmlstr += f"{str(m.visco2):^8s}  'visco2' = deformation-dependent Laplacian  viscosity factor\n"
    nmlstr += f"{str(m.visco4):^8s}  'visco4' = deformation-dependent biharmonic viscosity factor\n"
    nmlstr += f"{str(m.facdf4):^8s}  'facdf4' =       speed-dependent biharmonic viscosity factor\n"
    nmlstr += f"{str(m.veldf2):^8s}  'veldf2' = diffusion velocity (m/s) for Laplacian  momentum dissip.\n"
    nmlstr += f"{str(m.veldf4):^8s}  'veldf4' = diffusion velocity (m/s) for biharmonic momentum dissip.\n"
    nmlstr += f"{str(m.thkdf2):^8s}  'thkdf2' = diffusion velocity (m/s) for Laplacian  thickness diffus.\n"
    nmlstr += f"{str(m.thkdf4):^8s}  'thkdf4' = diffusion velocity (m/s) for biharmonic thickness diffus.\n"
    nmlstr += f"{str(m.temdf2):^8s}  'temdf2' = diffusion velocity (m/s) for Laplacian  temp/saln diffus.\n"
    nmlstr += f"{str(m.temdfc):^8s}  'temdfc' = temp diffusion conservation (0.0,1.0 all dens,temp resp.)\n"
    nmlstr += f"{str(m.vertmx):^8s}  'vertmx' = diffusion velocity (m/s) for momentum at MICOM M.L.base\n"
    nmlstr += f"{str(m.cbar  ):^8s}  'cbar  ' = rms flow speed     (m/s) for linear bottom friction\n"
    nmlstr += f"{str(m.cb    ):^8s}  'cb    ' = coefficient of quadratic bottom friction\n"
    nmlstr += f"{str(m.drglim):^8s}  'drglim' = limiter for explicit friction (1.0 none, 0.0 implicit)\n"
    nmlstr += f"{str(m.thkbot):^8s}  'thkbot' = thickness of bottom boundary layer (m)\n"
    nmlstr += f"{str(m.sigjmp):^8s}  'sigjmp' = minimum density jump across interfaces  (kg/m**3)\n"
    nmlstr += f"{str(m.tmljmp):^8s}  'tmljmp' = equivalent temperature jump across mixed-layer (degC)\n"
    nmlstr += f"{str(m.thkmls):^8s}  'thkmls' = reference mixed-layer thickness for SSS relaxation (m)\n"
    nmlstr += f"{str(m.thkmlt):^8s}  'thkmlt' = reference mixed-layer thickness for SST relaxation (m)\n"
    nmlstr += f"{str(m.thkriv):^8s}  'thkriv' = nominal thickness of river inflow (m)\n"
    nmlstr += f"{str(m.thkcdw):^8s}  'thkcdw' = thickness for near-surface currents in ice-ocean stress (m)\n"
    nmlstr += f"{str(m.thkfrz):^8s}  'thkfrz' = maximum thickness of near-surface freezing zone (m)\n"
    nmlstr += f"{str(m.iceflg):^8s}  'iceflg' = sea ice model flag (0=none,1=energy loan,2=coupled/esmf)\n"
    nmlstr += f"{str(m.tfrz_0):^8s}  'tfrz_0' = ENLN: ice melting point (degC) at S=0psu\n"
    nmlstr += f"{str(m.tfrz_s):^8s}  'tfrz_s' = ENLN: gradient of ice melting point (degC/psu)\n"
    nmlstr += f"{str(m.frzifq):^8s}  'frzifq' = e-folding time scale back to tfrz (days or -ve time steps)\n"
    nmlstr += f"{str(m.ticegr):^8s}  'ticegr' = ENLN: temp. grad. inside ice (deg/m); =0 use surtmp\n"
    nmlstr += f"{str(m.hicemn):^8s}  'hicemn' = ENLN: minimum ice thickness (m)\n"
    nmlstr += f"{str(m.hicemx):^8s}  'hicemx' = ENLN: maximum ice thickness (m)\n"
    nmlstr += f"{str(m.ishelf):^8s}  'ishelf' = ice shelf flag    (0=none,1=ice shelf over ocean)\n"
    nmlstr += f"{str(m.ntracr):^8s}  'ntracr' = number of tracers (0=none,negative to initialize)\n"
    nmlstr += f"{str(m.trcflg):^8s}  'trcflg' = tracer flags      (one digit per tr, most sig. replicated)\n"
    nmlstr += f"{str(m.tsofrq):^8s}  'tsofrq' = number of time steps between anti-drift offset calcs\n"
    nmlstr += f"{str(m.tofset):^8s}  'tofset' = temperature anti-drift offset (degC/century)\n"
    nmlstr += f"{str(m.sofset):^8s}  'sofset' = salnity     anti-drift offset  (psu/century)\n"
    nmlstr += f"{str(m.mlflag):^8s}  'mlflag' = mixed layer flag  (0=none,1=KPP,2-3=KT,4=PWP,5=MY,6=GISS)\n"
    nmlstr += f"{str(m.pensol):^8s}  'pensol' = KT:      activate penetrating solar rad.   (0=F,1=T)\n"
    nmlstr += f"{str(m.dtrate):^8s}  'dtrate' = KT:      maximum permitted m.l. detrainment rate  (m/day)\n"
    nmlstr += f"{str(m.thkmin):^8s}  'thkmin' = KT/PWP:  minimum mixed-layer thickness (m)\n"
    nmlstr += f"{str(m.dypflg):^8s}  'dypflg' = KT/PWP:  diapycnal mixing flag (0=none, 1=KPP, 2=explicit)\n"
    nmlstr += f"{str(m.mixfrq):^8s}  'mixfrq' = KT/PWP:  number of time steps between diapycnal mix calcs\n"
    nmlstr += f"{str(m.diapyc):^8s}  'diapyc' = KT/PWP:  diapycnal diffusivity x buoyancy freq. (m**2/s**2)\n"
    nmlstr += f"{str(m.rigr  ):^8s}  'rigr  ' = PWP:     critical gradient richardson number\n"
    nmlstr += f"{str(m.ribc  ):^8s}  'ribc  ' = PWP:     critical bulk     richardson number\n"
    nmlstr += f"{str(m.rinfty):^8s}  'rinfty' = KPP:     maximum  gradient richardson number (shear inst.)\n"
    nmlstr += f"{str(m.ricr  ):^8s}  'ricr  ' = KPP:     critical bulk     richardson number\n"
    nmlstr += f"{str(m.bldmin):^8s}  'bldmin' = KPP:     minimum surface boundary layer thickness (m)\n"
    nmlstr += f"{str(m.bldmax):^8s}  'bldmax' = K-PROF:  maximum surface boundary layer thickness (m)\n"
    nmlstr += f"{str(m.cekman):^8s}  'cekman' = KPP/KT:  scale factor for Ekman depth\n"
    nmlstr += f"{str(m.cmonob):^8s}  'cmonob' = KPP:     scale factor for Monin-Obukov depth\n"
    nmlstr += f"{str(m.bblkpp):^8s}  'bblkpp' = KPP:     activate bottom boundary layer    (0=F,1=T)\n"
    nmlstr += f"{str(m.shinst):^8s}  'shinst' = KPP:     activate shear instability mixing (0=F,1=T)\n"
    nmlstr += f"{str(m.dbdiff):^8s}  'dbdiff' = KPP:     activate double diffusion  mixing (0=F,1=T)\n"
    nmlstr += f"{str(m.nonloc):^8s}  'nonloc' = KPP:     activate nonlocal b. layer mixing (0=F,1=T)\n"
    nmlstr += f"{str(m.botdiw):^8s}  'botdiw' = K_PROF:    activate bot.enhan.int.wav mixing (0=F,1=T)\n"
    nmlstr += f"{str(m.difout):^8s}  'difout' = K-PROF:  output visc/diff coffs in archive (0=F,1=T)\n"
    nmlstr += f"{str(m.difsmo):^8s}  'difsmo' = K-PROF:  number of layers with horiz smooth diff coeffs\n"
    nmlstr += f"{str(m.difm0):^8s}  'difm0 ' = KPP:     max viscosity   due to shear instability (m**2/s)\n"
    nmlstr += f"{str(m.difs0):^8s}  'difs0 ' = KPP:     max diffusivity due to shear instability (m**2/s)\n"
    nmlstr += f"{str(m.difmiw):^8s}  'difmiw' = KPP:     background/internal wave viscosity       (m**2/s)\n"
    nmlstr += f"{str(m.difsiw):^8s}  'difsiw' = KPP:     background/internal wave diffusivity     (m**2/s)\n"
    nmlstr += f"{str(m.dsfmax):^8s}  'dsfmax' = KPP:     salt fingering diffusivity factor        (m**2/s)\n"
    nmlstr += f"{str(m.rrho0):^8s}  'rrho0 ' = KPP:     salt fingering rp=(alpha*delT)/(beta*delS)\n"
    nmlstr += f"{str(m.cs):^8s}  'cs    ' = KPP:     value for nonlocal flux term\n"
    nmlstr += f"{str(m.cstar):^8s}  'cstar ' = KPP:     value for nonlocal flux term\n"
    nmlstr += f"{str(m.cv):^8s}  'cv    ' = KPP:     buoyancy frequency ratio (0.0 to use a fn. of N)\n"
    nmlstr += f"{str(m.c11):^8s}  'c11   ' = KPP:     value for turb velocity scale\n"
    nmlstr += f"{str(m.hblflg):^8s}  'hblflg' = KPP:     b. layer interp. flag (0=const.,1=linear,2=quad.)\n"
    nmlstr += f"{str(m.niter):^8s}  'niter ' = KPP:     iterations for semi-implicit soln. (2 recomended)\n"
    nmlstr += f"{str(m.langmr):^8s}  'langmr' = KPP:     activate Langmuir turb. factor    (0=F,1=T)\n"
    nmlstr += f"{str(m.fltflg):^8s}  'fltflg' = FLOATS: synthetic float flag (0=no; 1=yes)\n"
    nmlstr += f"{str(m.nfladv):^8s}  'nfladv' = FLOATS: advect every nfladv bacl. time steps (even, >=4)\n"
    nmlstr += f"{str(m.nflsam):^8s}  'nflsam' = FLOATS: output (0=every nfladv steps; >0=no. of days)\n"
    nmlstr += f"{str(m.intpfl):^8s}  'intpfl' = FLOATS: horiz. interp. (0=2nd order+n.n.; 1=n.n. only)\n"
    nmlstr += f"{str(m.iturbv):^8s}  'iturbv' = FLOATS: add horiz. turb. advection velocity (0=no; 1=yes)\n"
    nmlstr += f"{str(m.ismpfl):^8s}  'ismpfl' = FLOATS: sample water properties at float (0=no; 1=yes)\n"
    nmlstr += f"{str(m.tbvar):^8s}  'tbvar ' = FLOATS: horizontal turb. vel. variance scale (m**2/s**2)\n"
    nmlstr += f"{str(m.tdecri):^8s}  'tdecri' = FLOATS: inverse decorrelation time scale (1/day)\n"
    nmlstr += f"{str(m.lbflag):^8s}  'lbflag' = lateral barotropic bndy flag (0=none, 1=port, 2=input)\n"
    nmlstr += f"{str(m.tidflg):^8s}  'tidflg' = TIDES: tidal forcing flag    (0=none,1=open-bdy,2=bdy&body)\n"
    nmlstr += f"{str(m.tidein):^8s}  'tidein' = TIDES: tide field input flag (0=no;1=yes;2=sal)\n"
    nmlstr += f"{str(m.tidcon):^8s}  'tidcon' = TIDES: 1 digit per (Q1K2P1N2O1K1S2M2), 0=off,1=on\n"
    nmlstr += f"{str(m.tidsal):^8s}  'tidsal' = TIDES: scalar self attraction and loading factor (<0: file)\n"
    nmlstr += f"{str(m.tiddrg):^8s}  'tiddrg' = TIDES: tidal drag flag (0=no;1=scalar;2=tensor)\n"
    nmlstr += f"{str(m.thkdrg):^8s}  'thkdrg' = thickness of bottom boundary layer for tidal drag (m)\n"
    nmlstr += f"{str(m.drgscl):^8s}  'drgscl' = scale factor for tidal drag   (0.0 for no tidal drag)\n"
    nmlstr += f"{str(m.tidgen):^8s}  'tidgen' = TIDES: generic time (0=F,1=T)\n"
    nmlstr += f"{str(m.tidrmp):^8s}  'tidrmp' = TIDES:            ramp time (days)\n"
    nmlstr += f"{str(m.tid_t0):^8s}  'tid_t0' = TIDES: origin for ramp time (model day)\n"
    nmlstr += f"{str(m.clmflg):^8s}  'clmflg' = climatology frequency flag   (6=bimonthly, 12=monthly)\n"
    nmlstr += f"{str(m.wndflg):^8s}  'wndflg' = wind stress input flag (0=none,1=u/v-grid,2,3=p-grid,4,5=wnd10m)\n"
    nmlstr += f"{str(m.ustflg):^8s}  'ustflg' = ustar forcing     flag        (3=input,1,2=wndspd,4=stress)\n"
    nmlstr += f"{str(m.flxflg):^8s}  'flxflg' = thermal forcing   flag (0=none,3=net-flux,1,2,4-6=sst-based)\n"
    nmlstr += f"{str(m.empflg):^8s}  'empflg' = E-P     forcing   flag (0=none,3=net_E-P, 1,2,4-6=sst-based_E)\n"
    nmlstr += f"{str(m.dswflg):^8s}  'dswflg' = diurnal shortwave flag (0=none,1=daily to diurnal corr.)\n"
    nmlstr += f"{str(m.albflg):^8s}  'albflg' = ocean albedo      flag (0=none,1=const,2=L&Y)\n"
    nmlstr += f"{str(m.sssflg):^8s}  'sssflg' = SSS relaxation flag (0=none,1=clim)\n"
    nmlstr += f"{str(m.lwflag):^8s}  'lwflag' = longwave (SST) flag (0=none,1=clim,2=atmos)\n"
    nmlstr += f"{str(m.sstflg):^8s}  'sstflg' = SST relaxation flag (0=none,1=clim,2=atmos,3=observed)\n"
    nmlstr += f"{str(m.icmflg):^8s}  'icmflg' = ice mask       flag (0=none,1=clim,2=atmos,3=obs/coupled)\n"
    nmlstr += f"{str(m.prsbas):^8s}  'prsbas' = msl pressure is input field + prsbas (Pa)\n"
    nmlstr += f"{str(m.mslprf):^8s}  'mslprf' = msl pressure forcing flag            (0=F,1=T)\n"
    nmlstr += f"{str(m.stroff):^8s}  'stroff' = net stress offset flag               (0=F,1=T)\n"
    nmlstr += f"{str(m.flxoff):^8s}  'flxoff' = net flux offset flag                 (0=F,1=T)\n"
    nmlstr += f"{str(m.flxsmo):^8s}  'flxsmo' = smooth surface fluxes                (0=F,1=T)\n"
    nmlstr += f"{str(m.relax):^8s}  'relax ' = activate lateral boundary nudging    (0=F,1=T)\n"
    nmlstr += f"{str(m.trcrlx):^8s}  'trcrlx' = activate lat. bound. tracer nudging  (0=F,1=T)\n"
    nmlstr += f"{str(m.priver):^8s}  'priver' = rivers as a precipitation bogas      (0=F,1=T)\n"
    nmlstr += f"{str(m.epmass):^8s}  'epmass' = treat evap-precip as a mass exchange (0=F,1=T)\n"
    nmlstr += f"{str(m.nmrsti):^8s}  'nmrsti' = 'restart'  Restart filename. Dot and date appended. Empty: use old hycom name.\n"
    nmlstr += f"{str(m.nmrsto):^8s}  'nmrsto' = 'restart'  Restart filename. Dot and date appended. Empty: use old hycom name.\n"
    nmlstr += f"{str(m.nmarcs):^8s}  'nmarcs' = '' Surface archive filename. Dot and date appended. Empty: use old hycom name.\n"
    nmlstr += f"{str(m.nmarcv):^8s}  'nmarcv' = '' Full    archive filename. Dot and date appended. Empty: use old hycom name.\n"
    nmlstr += f"{str(m.nmarcm):^8s}  'nmarcm' = '' Mean    archive filename. Dot and date appended. Empty: use old hycom name.\n"
    nmlstr += f"{str(m.stdflg):^8s}  'stdflg' = STOKES: add Stokes Drift velocities to kinematics and dynamics (0=F,1=T)\n"
    nmlstr += f"{str(m.stdsur):^8s}  'stdsur' = STOKES: add Stokes Drift Surface Stresses  to dnamics (0=F,1=T)\n"
    nmlstr += f"{str(m.stdtau):^8s}  'stdtau' = Remov from the tot wind stress the part lost to wave : (0=F,1=T)\n"
    nmlstr += f"{str(m.stdwom):^8s}  'stdwom' = Add wave-to-ocean momentum flux due to wave breaking  : (0=F,1=T)\n"
    nmlstr += f"{str(m.stdbot):^8s}  'stdbot' = STOKES: add Stokes Waves Field bottom friction drag (0=F,1=T)\n"
    nmlstr += f"{str(m.stdarc):^8s}  'stdarc' = STOKES: Stokes Drift Velocities in Archive (0=F,1=T)\n"
    nmlstr += f"{str(m.altlng):^8s}  'altlng' = STOKES: Use alternative Lang def. from Kai et al (0=F,1=T)\n"
    nmlstr += f"{str(m.nsdzi):^8s}  'nsdzi' = STOKES: number of interfaces in Stokes Drift input\n"
    return nmlstr

def limits(m, time, forecast_period):
    next_time = time + forecast_period * timedelta(hours=1)
    fdtime = dayfor(m.yrflag, time.year, time.timetuple().tm_yday, time.hour)
    ldtime = dayfor(m.yrflag, next_time.year, next_time.timetuple().tm_yday, next_time.hour)
    nmlstr = "{:14.5f} {:14.5f}".format(fdtime, ldtime)
    return nmlstr

def ports(m):
    nmlstr =  f"{str(m.nports):^8s} 'nports   ' = number of ports\n"
    for i in range(m.nports):
        nmlstr += f"{str(m.kdport[i]):^8s} 'kdport   ' = port orientation (1=N, 2=S, 3=E, 4=W)\n"
        nmlstr += f"{str(m.ifport[i]):^8s} 'ifport   ' = first i-index\n"
        nmlstr += f"{str(m.ilport[i]):^8s} 'ilport   ' = last  i-index (=ifport for east/west port)\n"
        nmlstr += f"{str(m.jfport[i]):^8s} 'jfport   ' = first j-index\n"
        nmlstr += f"{str(m.jlport[i]):^8s} 'jlport   ' = last  j-index (=jfport for north/south port)\n"
    return nmlstr

def ice_in(m, time, forecast_period):
    year_init = 1958 ##time.year
    time_init = datetime(year_init, 1, 1, 0, 0, 0, tzinfo=timezone.utc)
    sec0 = (time - time_init) / timedelta(seconds=1)
    istep0 = int(np.floor(sec0 / m.cice_dt)) - 1
    npt = int(np.floor(forecast_period * 3600 / m.cice_dt)) + 1

    ##namelist input for cice model "ice_in"
    nmlstr  = f"&setup_nml\n"
    nmlstr += f"    days_per_year = 365\n"
    nmlstr += f"    use_leap_years = .true.\n"
    nmlstr += f"    year_init = {year_init}\n"
    nmlstr += f"    istep0 = {istep0}\n"
    nmlstr += f"    dt = {m.cice_dt}\n"
    nmlstr += f"    npt = {npt}\n"
    nmlstr += f"    ndtd = 1\n"
    nmlstr += f"    runtype = 'continue'\n"
    nmlstr += f"    ice_ic = './cice/iced_gx1_v5.nc'\n"
    nmlstr += f"    restart = .true.\n"
    nmlstr += f"    restart_ext = .false.\n"
    nmlstr += f"    use_restart_time = .true.\n"
    nmlstr += f"    restart_format = 'nc'\n"
    nmlstr += f"    lcdf64 = .false.\n"
    nmlstr += f"    restart_dir = './cice/'\n"
    nmlstr += f"    restart_file = 'iced'\n"
    nmlstr += f"    pointer_file = './cice/ice.restart_file'\n"
    nmlstr += f"    dumpfreq = 'd'\n"
    nmlstr += f"    dumpfreq_n = 60\n"
    nmlstr += f"    dump_last = .true.\n"
    nmlstr += f"    bfbflag = .false.\n"
    nmlstr += f"    diagfreq = 24\n"
    nmlstr += f"    diag_type = 'stdout'\n"
    nmlstr += f"    diag_file = 'ice_diag.d'\n"
    nmlstr += f"    print_global = .false.\n"
    nmlstr += f"    print_points = .false.\n"
    nmlstr += f"    latpnt(1:2) = 90.0, 75.0\n"
    nmlstr += f"    lonpnt(1:2) = 0.0, 145.0\n"
    nmlstr += f"    dbug = .false.\n"
    nmlstr += f"    histfreq = 'h', 'd', 'm', 'x', 'x'\n"
    nmlstr += f"    histfreq_n = 0, 1, 0, 1, 1\n"
    nmlstr += f"    hist_avg = .true.\n"
    nmlstr += f"    history_dir = './cice/'\n"
    nmlstr += f"    history_file = 'iceh'\n"
    nmlstr += f"    write_ic = .true.\n"
    nmlstr += f"    incond_dir = './cice/'\n"
    nmlstr += f"    incond_file = 'iceh_ic'\n"
    nmlstr += f"/\n"
    nmlstr += f"\n"
    nmlstr += f"&grid_nml\n"
    nmlstr += f"    grid_format = 'nc'\n"
    nmlstr += f"    grid_type = 'regional'\n"
    nmlstr += f"    grid_file = 'cice_grid.nc'\n"
    nmlstr += f"    kmt_file = 'cice_kmt.nc'\n"
    nmlstr += f"    gridcpl_file = 'unknown_gridcpl_file'\n"
    nmlstr += f"    kcatbound = 0\n"
    nmlstr += f"/\n"
    nmlstr += f"\n"
    nmlstr += f"&domain_nml\n"
    nmlstr += f"    nprocs = {m.nproc}\n"
    nmlstr += f"    processor_shape = 'square-pop'\n"
    nmlstr += f"    distribution_type = 'cartesian'\n"
    nmlstr += f"    distribution_wght = 'block'\n"
    nmlstr += f"    ew_boundary_type = 'open'\n"
    nmlstr += f"    ns_boundary_type = 'open'\n"
    nmlstr += f"    maskhalo_dyn = .false.\n"
    nmlstr += f"    maskhalo_remap = .false.\n"
    nmlstr += f"    maskhalo_bound = .false.\n"
    nmlstr += f"/\n"
    nmlstr += f"\n"
    nmlstr += f"&tracer_nml\n"
    nmlstr += f"    tr_iage = .true.\n"
    nmlstr += f"    restart_age = .false.\n"
    nmlstr += f"    tr_fy = .true.\n"
    nmlstr += f"    restart_fy = .false.\n"
    nmlstr += f"    tr_lvl = .true.\n"
    nmlstr += f"    restart_lvl = .false.\n"
    nmlstr += f"    tr_pond_cesm = .false.\n"
    nmlstr += f"    restart_pond_cesm = .false.\n"
    nmlstr += f"    tr_pond_topo = .false.\n"
    nmlstr += f"    restart_pond_topo = .false.\n"
    nmlstr += f"    tr_pond_lvl = .true.\n"
    nmlstr += f"    restart_pond_lvl = .false.\n"
    nmlstr += f"    tr_aero = .false.\n"
    nmlstr += f"    restart_aero = .false.\n"
    nmlstr += f"/\n"
    nmlstr += f"\n"
    nmlstr += f"&thermo_nml\n"
    nmlstr += f"    kitd = 1\n"
    nmlstr += f"    ktherm = 1\n"
    nmlstr += f"    conduct = 'bubbly'\n"
    nmlstr += f"    a_rapid_mode = 0.0005\n"
    nmlstr += f"    rac_rapid_mode = 10.0\n"
    nmlstr += f"    aspect_rapid_mode = 1.0\n"
    nmlstr += f"    dsdt_slow_mode = -5e-08\n"
    nmlstr += f"    phi_c_slow_mode = 0.05\n"
    nmlstr += f"    phi_i_mushy = 0.85\n"
    nmlstr += f"/\n"
    nmlstr += f"\n"
    nmlstr += f"&dynamics_nml\n"
    nmlstr += f"    kdyn = 1\n"
    nmlstr += f"    ndte = 120\n"
    nmlstr += f"    revised_evp = .false.\n"
    nmlstr += f"    advection = 'remap'\n"
    nmlstr += f"    kstrength = 0\n"
    nmlstr += f"    krdg_partic = 1\n"
    nmlstr += f"    krdg_redist = 1\n"
    nmlstr += f"    mu_rdg = 3\n"
    nmlstr += f"    cf = 17.0\n"
    nmlstr += f"/\n"
    nmlstr += f"\n"
    nmlstr += f"&shortwave_nml\n"
    nmlstr += f"    shortwave = 'dEdd'\n"
    nmlstr += f"    albedo_type = 'constant'\n"
    nmlstr += f"    albicev = 0.78\n"
    nmlstr += f"    albicei = 0.36\n"
    nmlstr += f"    albsnowv = 0.98\n"
    nmlstr += f"    albsnowi = 0.7\n"
    nmlstr += f"    ahmax = 0.3\n"
    nmlstr += f"    r_ice = 1.5\n"
    nmlstr += f"    r_pnd = 1.2\n"
    nmlstr += f"    r_snw = 1.2\n"
    nmlstr += f"    dt_mlt = 1.5\n"
    nmlstr += f"    rsnw_mlt = 1500.0\n"
    nmlstr += f"    kalg = 0.6\n"
    nmlstr += f"/\n"
    nmlstr += f"\n"
    nmlstr += f"&ponds_nml\n"
    nmlstr += f"    hp1 = 0.01\n"
    nmlstr += f"    hs0 = 0.0\n"
    nmlstr += f"    hs1 = 0.03\n"
    nmlstr += f"    dpscale = 0.001\n"
    nmlstr += f"    frzpnd = 'hlid'\n"
    nmlstr += f"    rfracmin = 0.15\n"
    nmlstr += f"    rfracmax = 1.0\n"
    nmlstr += f"    pndaspect = 0.8\n"
    nmlstr += f"/\n"
    nmlstr += f"\n"
    nmlstr += f"&zbgc_nml\n"
    nmlstr += f"    tr_brine = .false.\n"
    nmlstr += f"    restart_hbrine = .false.\n"
    nmlstr += f"    skl_bgc = .false.\n"
    nmlstr += f"    bgc_flux_type = 'Jin2006'\n"
    nmlstr += f"    restart_bgc = .false.\n"
    nmlstr += f"    restore_bgc = .false.\n"
    nmlstr += f"    bgc_data_dir = 'unknown_bgc_data_dir'\n"
    nmlstr += f"    sil_data_type = 'default'\n"
    nmlstr += f"    nit_data_type = 'default'\n"
    nmlstr += f"    tr_bgc_c_sk = .false.\n"
    nmlstr += f"    tr_bgc_chl_sk = .false.\n"
    nmlstr += f"    tr_bgc_am_sk = .false.\n"
    nmlstr += f"    tr_bgc_sil_sk = .false.\n"
    nmlstr += f"    tr_bgc_dmspp_sk = .false.\n"
    nmlstr += f"    tr_bgc_dmspd_sk = .false.\n"
    nmlstr += f"    tr_bgc_dms_sk = .false.\n"
    nmlstr += f"    phi_snow = 0.5\n"
    nmlstr += f"/\n"
    nmlstr += f"\n"
    nmlstr += f"&forcing_nml\n"
    nmlstr += f"    formdrag = .false.\n"
    nmlstr += f"    atmbndy = 'default'\n"
    nmlstr += f"    fyear_init = {year_init}\n"
    nmlstr += f"    ycycle = 52\n"
    nmlstr += f"    atm_data_format = 'bin'\n"
    nmlstr += f"    atm_data_type = 'None'\n"
    nmlstr += f"    atm_data_dir = '/usr/projects/climate/eclare/DATA/gx1v3/LargeYeager/v2_updated/'\n"
    nmlstr += f"    calc_strair = .true.\n"
    nmlstr += f"    highfreq = .false.\n"
    nmlstr += f"    natmiter = 5\n"
    nmlstr += f"    calc_tsfc = .true.\n"
    nmlstr += f"    precip_units = 'mm_per_sec'\n"
    nmlstr += f"    ustar_min = 0.0005\n"
    nmlstr += f"    fbot_xfer_type = 'constant'\n"
    nmlstr += f"    update_ocn_f = .true.\n"
    nmlstr += f"    l_mpond_fresh = .false.\n"
    nmlstr += f"    tfrz_option = 'linear_salt'\n"
    nmlstr += f"    oceanmixed_ice = .false.\n"
    nmlstr += f"    ocn_data_format = 'nc'\n"
    nmlstr += f"    sss_data_type = 'default'\n"
    nmlstr += f"    sst_data_type = 'default'\n"
    nmlstr += f"    ocn_data_dir = '/usr/projects/climate/eclare/DATA/gx1v3/gx1v3/forcing/'\n"
    nmlstr += f"    oceanmixed_file = 'oceanmixed_ice_depth.nc'\n"
    nmlstr += f"    restore_sst = .false.\n"
    nmlstr += f"    trestore = 90\n"
    nmlstr += f"    restore_ice = .false.\n"
    nmlstr += f"/\n"
    nmlstr += f"\n"
    nmlstr += f"&icefields_nml\n"
    nmlstr += f"    f_tmask = .true.\n"
    nmlstr += f"    f_blkmask = .true.\n"
    nmlstr += f"    f_tarea = .true.\n"
    nmlstr += f"    f_uarea = .true.\n"
    nmlstr += f"    f_dxt = .false.\n"
    nmlstr += f"    f_dyt = .false.\n"
    nmlstr += f"    f_dxu = .false.\n"
    nmlstr += f"    f_dyu = .false.\n"
    nmlstr += f"    f_htn = .false.\n"
    nmlstr += f"    f_hte = .false.\n"
    nmlstr += f"    f_angle = .true.\n"
    nmlstr += f"    f_anglet = .true.\n"
    nmlstr += f"    f_ncat = .true.\n"
    nmlstr += f"    f_vgrdi = .false.\n"
    nmlstr += f"    f_vgrds = .false.\n"
    nmlstr += f"    f_vgrdb = .false.\n"
    nmlstr += f"    f_bounds = .false.\n"
    nmlstr += f"    f_aice = 'd'\n"
    nmlstr += f"    f_hi = 'd'\n"
    nmlstr += f"    f_hs = 'd'\n"
    nmlstr += f"    f_tsfc = 'd'\n"
    nmlstr += f"    f_sice = 'd'\n"
    nmlstr += f"    f_uvel = 'd'\n"
    nmlstr += f"    f_vvel = 'd'\n"
    nmlstr += f"    f_uatm = 'h'\n"
    nmlstr += f"    f_vatm = 'h'\n"
    nmlstr += f"    f_fswdn = 'h'\n"
    nmlstr += f"    f_flwdn = 'h'\n"
    nmlstr += f"    f_snow = 'd'\n"
    nmlstr += f"    f_snow_ai = 'm'\n"
    nmlstr += f"    f_rain = 'h'\n"
    nmlstr += f"    f_rain_ai = 'm'\n"
    nmlstr += f"    f_sst = 'd'\n"
    nmlstr += f"    f_sss = 'd'\n"
    nmlstr += f"    f_uocn = 'd'\n"
    nmlstr += f"    f_vocn = 'd'\n"
    nmlstr += f"    f_frzmlt = 'h'\n"
    nmlstr += f"    f_fswfac = 'm'\n"
    nmlstr += f"    f_fswint_ai = 'm'\n"
    nmlstr += f"    f_fswabs = 'x'\n"
    nmlstr += f"    f_fswabs_ai = 'm'\n"
    nmlstr += f"    f_albsni = 'm'\n"
    nmlstr += f"    f_alvdr = 'x'\n"
    nmlstr += f"    f_alidr = 'x'\n"
    nmlstr += f"    f_alvdf = 'x'\n"
    nmlstr += f"    f_alidf = 'x'\n"
    nmlstr += f"    f_albice = 'd'\n"
    nmlstr += f"    f_albsno = 'd'\n"
    nmlstr += f"    f_albpnd = 'd'\n"
    nmlstr += f"    f_coszen = 'x'\n"
    nmlstr += f"    f_flat = 'x'\n"
    nmlstr += f"    f_flat_ai = 'm'\n"
    nmlstr += f"    f_fsens = 'x'\n"
    nmlstr += f"    f_fsens_ai = 'm'\n"
    nmlstr += f"    f_flwup = 'h'\n"
    nmlstr += f"    f_flwup_ai = 'h'\n"
    nmlstr += f"    f_evap = 'x'\n"
    nmlstr += f"    f_evap_ai = 'm'\n"
    nmlstr += f"    f_tair = 'h'\n"
    nmlstr += f"    f_tref = 'x'\n"
    nmlstr += f"    f_qref = 'x'\n"
    nmlstr += f"    f_congel = 'm'\n"
    nmlstr += f"    f_frazil = 'm'\n"
    nmlstr += f"    f_snoice = 'm'\n"
    nmlstr += f"    f_dsnow = 'x'\n"
    nmlstr += f"    f_melts = 'm'\n"
    nmlstr += f"    f_meltt = 'm'\n"
    nmlstr += f"    f_meltb = 'm'\n"
    nmlstr += f"    f_meltl = 'm'\n"
    nmlstr += f"    f_fresh = 'h'\n"
    nmlstr += f"    f_fresh_ai = 'h'\n"
    nmlstr += f"    f_fsalt = 'h'\n"
    nmlstr += f"    f_fsalt_ai = 'h'\n"
    nmlstr += f"    f_fhocn = 'h'\n"
    nmlstr += f"    f_fhocn_ai = 'h'\n"
    nmlstr += f"    f_fswthru = 'h'\n"
    nmlstr += f"    f_fswthru_ai = 'h'\n"
    nmlstr += f"    f_fsurf_ai = 'h'\n"
    nmlstr += f"    f_fcondtop_ai = 'x'\n"
    nmlstr += f"    f_fmeltt_ai = 'h'\n"
    nmlstr += f"    f_strairx = 'h'\n"
    nmlstr += f"    f_strairy = 'h'\n"
    nmlstr += f"    f_strtltx = 'h'\n"
    nmlstr += f"    f_strtlty = 'h'\n"
    nmlstr += f"    f_strcorx = 'x'\n"
    nmlstr += f"    f_strcory = 'x'\n"
    nmlstr += f"    f_strocnx = 'h'\n"
    nmlstr += f"    f_strocny = 'h'\n"
    nmlstr += f"    f_strintx = 'x'\n"
    nmlstr += f"    f_strinty = 'x'\n"
    nmlstr += f"    f_strength = 'm'\n"
    nmlstr += f"    f_divu = 'm'\n"
    nmlstr += f"    f_shear = 'm'\n"
    nmlstr += f"    f_sig1 = 'm'\n"
    nmlstr += f"    f_sig2 = 'm'\n"
    nmlstr += f"    f_dvidtt = 'm'\n"
    nmlstr += f"    f_dvidtd = 'm'\n"
    nmlstr += f"    f_daidtt = 'm'\n"
    nmlstr += f"    f_daidtd = 'm'\n"
    nmlstr += f"    f_dagedtt = 'm'\n"
    nmlstr += f"    f_dagedtd = 'm'\n"
    nmlstr += f"    f_mlt_onset = 'm'\n"
    nmlstr += f"    f_frz_onset = 'm'\n"
    nmlstr += f"    f_hisnap = 'x'\n"
    nmlstr += f"    f_aisnap = 'x'\n"
    nmlstr += f"    f_trsig = 'm'\n"
    nmlstr += f"    f_icepresent = 'h'\n"
    nmlstr += f"    f_iage = 'm'\n"
    nmlstr += f"    f_fy = 'm'\n"
    nmlstr += f"    f_aicen = 'x'\n"
    nmlstr += f"    f_vicen = 'x'\n"
    nmlstr += f"    f_vsnon = 'x'\n"
    nmlstr += f"    f_keffn_top = 'x'\n"
    nmlstr += f"    f_tinz = 'x'\n"
    nmlstr += f"    f_sinz = 'x'\n"
    nmlstr += f"    f_tsnz = 'x'\n"
    nmlstr += f"    f_fsurfn_ai = 'h'\n"
    nmlstr += f"    f_fcondtopn_ai = 'h'\n"
    nmlstr += f"    f_fmelttn_ai = 'h'\n"
    nmlstr += f"    f_flatn_ai = 'x'\n"
    nmlstr += f"    f_fsensn_ai = 'x'\n"
    nmlstr += f"/\n"
    nmlstr += f"\n"
    nmlstr += f"&icefields_mechred_nml\n"
    nmlstr += f"    f_alvl = 'm'\n"
    nmlstr += f"    f_vlvl = 'm'\n"
    nmlstr += f"    f_ardg = 'm'\n"
    nmlstr += f"    f_vrdg = 'm'\n"
    nmlstr += f"    f_dardg1dt = 'x'\n"
    nmlstr += f"    f_dardg2dt = 'x'\n"
    nmlstr += f"    f_dvirdgdt = 'x'\n"
    nmlstr += f"    f_opening = 'x'\n"
    nmlstr += f"    f_ardgn = 'x'\n"
    nmlstr += f"    f_vrdgn = 'x'\n"
    nmlstr += f"    f_dardg1ndt = 'x'\n"
    nmlstr += f"    f_dardg2ndt = 'x'\n"
    nmlstr += f"    f_dvirdgndt = 'x'\n"
    nmlstr += f"    f_krdgn = 'x'\n"
    nmlstr += f"    f_aparticn = 'x'\n"
    nmlstr += f"    f_aredistn = 'x'\n"
    nmlstr += f"    f_vredistn = 'x'\n"
    nmlstr += f"    f_araftn = 'x'\n"
    nmlstr += f"    f_vraftn = 'x'\n"
    nmlstr += f"/\n"
    nmlstr += f"\n"
    nmlstr += f"&icefields_pond_nml\n"
    nmlstr += f"    f_apondn = 'x'\n"
    nmlstr += f"    f_apeffn = 'x'\n"
    nmlstr += f"    f_hpondn = 'x'\n"
    nmlstr += f"    f_apond = 'm'\n"
    nmlstr += f"    f_hpond = 'm'\n"
    nmlstr += f"    f_ipond = 'm'\n"
    nmlstr += f"    f_apeff = 'm'\n"
    nmlstr += f"    f_apond_ai = 'm'\n"
    nmlstr += f"    f_hpond_ai = 'm'\n"
    nmlstr += f"    f_ipond_ai = 'm'\n"
    nmlstr += f"    f_apeff_ai = 'm'\n"
    nmlstr += f"/\n"
    nmlstr += f"\n"
    nmlstr += f"&icefields_bgc_nml\n"
    nmlstr += f"    f_faero_atm = 'x'\n"
    nmlstr += f"    f_faero_ocn = 'x'\n"
    nmlstr += f"    f_aero = 'x'\n"
    nmlstr += f"    f_fno = 'x'\n"
    nmlstr += f"    f_fno_ai = 'x'\n"
    nmlstr += f"    f_fnh = 'x'\n"
    nmlstr += f"    f_fnh_ai = 'x'\n"
    nmlstr += f"    f_fn = 'x'\n"
    nmlstr += f"    f_fn_ai = 'x'\n"
    nmlstr += f"    f_fsil = 'x'\n"
    nmlstr += f"    f_fsil_ai = 'x'\n"
    nmlstr += f"    f_bgc_n_sk = 'x'\n"
    nmlstr += f"    f_bgc_c_sk = 'x'\n"
    nmlstr += f"    f_bgc_chl_sk = 'x'\n"
    nmlstr += f"    f_bgc_nit_sk = 'x'\n"
    nmlstr += f"    f_bgc_am_sk = 'x'\n"
    nmlstr += f"    f_bgc_sil_sk = 'x'\n"
    nmlstr += f"    f_bgc_dmspp_sk = 'x'\n"
    nmlstr += f"    f_bgc_dmspd_sk = 'x'\n"
    nmlstr += f"    f_bgc_dms_sk = 'x'\n"
    nmlstr += f"    f_bgc_nit_ml = 'x'\n"
    nmlstr += f"    f_bgc_am_ml = 'x'\n"
    nmlstr += f"    f_bgc_sil_ml = 'x'\n"
    nmlstr += f"    f_bgc_dmsp_ml = 'x'\n"
    nmlstr += f"    f_btin = 'x'\n"
    nmlstr += f"    f_bphi = 'x'\n"
    nmlstr += f"    f_fbri = 'm'\n"
    nmlstr += f"    f_hbri = 'm'\n"
    nmlstr += f"    f_grownet = 'x'\n"
    nmlstr += f"    f_ppnet = 'x'\n"
    nmlstr += f"/\n"
    nmlstr += f"\n"
    nmlstr += f"&icefields_drag_nml\n"
    nmlstr += f"    f_drag = 'x'\n"
    nmlstr += f"    f_cdn_atm = 'x'\n"
    nmlstr += f"    f_cdn_ocn = 'x'\n"
    nmlstr += f"/\n"
    return nmlstr

def namelist(m, time, forecast_period, run_dir='.'):
    """
    Generate namelists for TP5 model runs,
    Inputs:
    - m: Model object with configs
    - run_dir: path to the run directory
    """
    with open(os.path.join(run_dir, 'blkdat.input'), 'wt') as f:
        f.write(blkdat(m))

    with open(os.path.join(run_dir, 'limits'), 'wt') as f:
        f.write(limits(m, time, forecast_period))

    with open(os.path.join(run_dir, 'ports.input'), 'wt') as f:
        f.write(ports(m))

    with open(os.path.join(run_dir, 'ice_in'), 'wt') as f:
        f.write(ice_in(m, time, forecast_period))




import os
import subprocess
import numpy as np
from pyproj import Proj
from NEDAS.grid import Grid
from NEDAS.utils.conversion import units_convert, t2s, dt1h
from NEDAS.utils.shell_utils import run_job, makedir
from NEDAS.utils.netcdf_lib import nc_read_var, nc_write_var
from NEDAS.models import Model
from . import restart, forcing, namelist, dgLimit

class NextsimDGModel(Model):
    restart_dt: float
    dg_comp: int

    proj: str
    xstart: float
    xend: float
    ystart: float
    yend: float
    dx: float
    mask_file: str

    files: dict
    perturb: dict

    model_env: str
    model_config_file: str
    nproc_per_run: int
    walltime: int
    parallel_mode: str
    run_separate_jobs: bool
    use_job_array: bool

    def __init__(self, config_file=None, parse_args=False, **kwargs):
        super().__init__(config_file, parse_args, **kwargs)

        self.native_variables = {
             'seaice_conc_dg': {'name':'data/cice', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':np.arange(self.dg_comp), 'units':1},
             'seaice_thick_dg': {'name':'data/hice', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':np.arange(self.dg_comp), 'units':'m'},
             'seaice_damage': {'name':'data/damage', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':1},
             'snow_thick': {'name':'data/hsnow', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':'m'},
             'seaice_temp_k': {'name':'data/tice', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0,1,2], 'units':'K'},
             'seaice_velocity': {'name':('data/u', 'data/v'), 'dtype':'float', 'is_vector':True, 'dt':self.restart_dt, 'levels':[0], 'units':'m/s'},
             }
        self.diag_variables = {
             'seaice_conc': {'name':'data/sic', 'operator':self.get_seaice_conc, 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':1},
             'seaice_thick': {'name':'data/sit', 'operator':self.get_seaice_thick, 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':'m'},
             }
        self.atmos_forcing_variables = {
            'atmos_surf_velocity': {'name':('data/u', 'data/v'), 'dtype':'float', 'is_vector':True, 'dt':self.restart_dt, 'levels':[0], 'units':'m/s'},
            'atmos_surf_press': {'name':'data/pair', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':'Pa'},
            'atmos_surf_temp': {'name':'data/tair', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':'C'},
            'atmos_surf_dewpoint': {'name':'data/dew2m', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':'C'},
            'atmos_down_shortwave': {'name':'data/sw_in', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':'W/m2'},
            'atmos_down_longwave': {'name':'data/lw_in', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':'W/m2'},
        }
        self.ocean_forcing_variables = {
            'ocean_surf_velocity': {'name':('data/u', 'data/v'), 'dtype':'float', 'is_vector':True, 'dt':self.restart_dt, 'levels':[0], 'units':'m/s'},
            'ocean_surf_temp': {'name':'data/sst', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':'Pa'},
            'ocean_surf_salinity': {'name':'data/sss', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':'Pa'},
            'ocean_mixl_depth': {'name':'data/mld', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':'Pa'},
        }
        self.variables = {**self.native_variables, **self.diag_variables, **self.atmos_forcing_variables, **self.ocean_forcing_variables}

        self.z_units = 'm'

        # construct grid obj based on config
        self.grid = Grid.regular_grid(Proj(self.proj), self.xstart, self.xend, self.ystart, self.yend, self.dx)

        # mask for grid points not
        self.grid.mask = self.read_mask()

    def filename(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)

        member = kwargs['member']
        if member is None:
            ens_mem_dir = ''
            ens_mem_id = 1
        else:
            ens_mem_id = member + 1
            ens_mem_dir = f'ens_{str(ens_mem_id).zfill(2)}'

        time = kwargs['time']
        assert time is not None, 'nextsim.dg.filename: time needs to be specified, wildcard searching is not implemented.'

        name = kwargs['name']  ##name of the variable
        if name in self.native_variables:
            fname = restart.get_restart_filename(self.files['restart'], ens_mem_id, time)

        elif name in self.diag_variables:
            fname = f"{name}_k{kwargs['k']}_{t2s(kwargs['time'])}.npy"

        elif name in self.atmos_forcing_variables:
            fname = forcing.get_forcing_filename(self.files['forcing']['atmosphere'], ens_mem_id, time)

        elif name in self.ocean_forcing_variables:
            fname = forcing.get_forcing_filename(self.files['forcing']['ocean'], ens_mem_id, time)

        else:
            raise ValueError(f"variable {name} is not defined for nextsimdg model.")

        ##fname is given by the format defined in config file
        ##the source file is copied to path (cycle directory) in preprocess
        ##this filename function will return the copy, not the original location defined by fname
        return os.path.join(kwargs['path'], ens_mem_dir, os.path.basename(fname))

    def read_grid(self, **kwargs):
        ## the nextsim dg grid is fixed, no need to update the grid at runtime
        pass

    def read_mask(self):
        ## the nextsim dg grid is fixed, no need to update the mask at runtime
        if os.path.exists(self.mask_file):
            mask = nc_read_var(self.mask_file, 'data/mask')[:]
            return (mask==0)
        return np.full(self.grid.x.shape, False)

    def read_var(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        fname = self.filename(**kwargs)
        name = kwargs['name']
        rec = self.variables[name]

        if name in self.native_variables:
            if rec['is_vector']:
                u = nc_read_var(fname, rec['name'][0])
                v = nc_read_var(fname, rec['name'][1])
                var = np.array([u, v])
            else:
                if rec['name'] in ['data/cice', 'data/hice', 'data/damage']:
                    tmp = nc_read_var(fname, rec['name'])
                    if len(tmp.shape) == 3:
                        var = tmp[..., kwargs['k']]
                    else:
                        var = tmp
                elif rec['name'] in ['data/tice']:
                    var = nc_read_var(fname, rec['name'])[kwargs['k'], ...]
                else:
                    var = nc_read_var(fname, rec['name'])

        elif name in self.diag_variables:
            var = rec['operator'](**kwargs)
            np.save(fname, var)

        else:
            if name in self.atmos_forcing_variables:
                forcing_name = 'atmosphere'
            elif name in self.ocean_forcing_variables:
                forcing_name = 'ocean'
            else:
                raise KeyError(f"variable {name} is not a forcing variable")
            file_options = self.files['forcing'][forcing_name]
            itime = forcing.get_time_index(fname,
                                           file_options['time_name'],
                                           file_options['time_units_name'],
                                           kwargs['time']
                                           )
            if rec['is_vector']:
                u = forcing.read_var(fname, [rec['name'][0],], itime)[0,...].data
                v = forcing.read_var(fname, [rec['name'][1],], itime)[0,...].data
                var = np.array([u, v])
            else:
                var = forcing.read_var(fname, [rec['name'],], itime)[0,...].data

        var = units_convert(rec['units'], kwargs['units'], var)
        return var

    def write_var(self, var, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        fname = self.filename(**kwargs)
        name = kwargs['name']
        rec = self.variables[name]

        var = units_convert(kwargs['units'], rec['units'], var)

        if name in self.native_variables:
            if rec['is_vector']:
                dims = {'ydim':self.grid.ny, 'xdim':self.grid.nx}
                for i in range(2):
                    nc_write_var(fname, dims, rec['name'][i], var[i,...], comm=kwargs['comm'])
            else:
                if rec['name'] in ['data/cice', 'data/hice']:
                    dims = {'ydim':self.grid.ny, 'xdim':self.grid.nx, 'dg_comp':None}
                    recno = {'dg_comp':kwargs['k']}
                    nc_write_var(fname, dims, rec['name'], var, recno=recno, comm=kwargs['comm'])
                elif rec['name'] in ['data/tice']:
                    dims = {'zdim':None, 'ydim':self.grid.ny, 'xdim':self.grid.nx}
                    recno = {'zdim':kwargs['k']}
                    nc_write_var(fname, dims, rec['name'], var, recno=recno, comm=kwargs['comm'])
                else:
                    dims = {'ydim':self.grid.ny, 'xdim':self.grid.nx}
                    nc_write_var(fname, dims, rec['name'], var, comm=kwargs['comm'])

        elif name in self.diag_variables:
            np.save(fname, var)

        else:
            if name in self.atmos_forcing_variables:
                forcing_name = 'atmosphere'
            elif name in self.ocean_forcing_variables:
                forcing_name = 'ocean'
            else:
                raise KeyError(f"variable {name} is not a forcing variable")
            file_options = self.files['forcing'][forcing_name]
            itime = forcing.get_time_index(fname,
                                           file_options['time_name'],
                                           file_options['time_units_name'],
                                           kwargs['time']
                                           )
            if rec['is_vector']:
                forcing.write_var(fname, list(rec['name']), np.ma.array(var), itime)
            else:
                forcing.write_var(fname, [rec['name'],], np.ma.array([var,]), itime)

    def z_coords(self, **kwargs):
        return np.zeros(self.grid.x.shape)

    def get_seaice_conc(self, **kwargs):
        return self.read_var(**{**kwargs, 'name':'seaice_conc_dg', 'k':0, 'units':1})

    def get_seaice_thick(self, **kwargs):
        return self.read_var(**{**kwargs, 'name':'seaice_thick_dg', 'k':0, 'units':'m'})

    def preprocess(self, task_id:int=0, **kwargs):
        """Preprocessing method for nextsim.dg
        Parameters
        ----------
        task_id : int
            task id for parallel execution
        self.nproc_per_util : int
            number of processors for each task
        **kwargs : dict
            keyword arguments for the model configuration
            Keywords defined when the function is called:
            - member : int
                ensemble member id
            - time : datetime
                start time of the forecast
            - time_start: datetime
                initial time of the forecast cycles
            - path : str
                path to the working directory of the ensemble member
            - forecast_period : int
                number of hours being forecasted by model
            - restart_dir : str
                the saved restart directory from the previous cycle,
                which is the model run directory from
                the previous cycle. In the initial cycle, this
                directory is given as `ens_init_dir` defined in
                `nextsim.dg`` section of the `model_def` section
                in nedas config file.
        These are defined in the `config_file` entry of model_def:nextsim.dg section,
        and parse_config will bring them in this class as
            - self.files : dict
                This section contains the filenames for the restart file.
                This must have a `restart` key that defines the filename `format`
                and the strftime `time_format` code of the restart filename.
                If perturbation is used, this section must also have a `lon_name`
                and `lat_name` key
            - self.perturb : dict
                perturbation options for the initial conditions.
                See example configuration file for required keys and explanations.
                This section is not necessary if the model does not use perturbation
                for the initial conditions or forcings.
        """
        kwargs = super().parse_kwargs(**kwargs)

        # get the current ensemble member id
        ens_mem_id:int = kwargs['member'] + 1
        # ensemble member directory for the current member
        ens_mem_dir:str = f'ens_{str(ens_mem_id).zfill(2)}'

        # directory where files are being collected to, and where the model will be run
        run_dir = os.path.join(kwargs['path'], ens_mem_dir)
        makedir(run_dir)

        # get all required filenames for the initial ensemble
        # 1. get current and next time
        time = kwargs['time']
        forecast_period = kwargs['forecast_period']
        next_time = time + forecast_period * dt1h
        time_start = kwargs['time_start']
        debug = kwargs['debug']

        # 2. get the restart and forcing filename
        file_options_restart = self.files['restart']
        # obtain restart file at initial cycling
        fname_restart_init:str = restart.get_restart_filename(file_options_restart, ens_mem_id, time)
        fname_restart:str = os.path.join(kwargs['restart_dir'], ens_mem_dir, os.path.basename(fname_restart_init))
        if not os.path.exists(fname_restart):
            fname_restart = fname_restart_init

        file_options_forcing:dict[str, dict] = self.files['forcing']
        fname_forcing:dict[str, str] = dict()
        for forcing_name in file_options_forcing:
            fname_forcing[forcing_name] = forcing.get_forcing_filename(file_options_forcing[forcing_name], ens_mem_id, time)

        # no need for perturbation if not specified in yaml file
        if not hasattr(self, 'perturb'):
            print('We do no perturbations as perturb section '
                  'is not specified in the model configuration.',
                  flush=True)
            # we we do not perturb the restart file
            # simply link the restart files
            os.system(f'ln -fs {fname_restart} {run_dir}')
            # we we do not perturb the forcing file
            # simply link the forcing files
            for forcing_name in file_options_forcing:
                os.system(f'ln -fs {fname_forcing[forcing_name]} {run_dir}')
            return

        # 3. add perturbations
        np.random.seed(time.second+ens_mem_id)
        # here, if 'restart section is not under perturb section
        # we only link the restart file to each ensemble directory
        if 'restart' not in self.perturb or kwargs['time'] != kwargs['time_start']:
            # we we do not perturb the restart file
            # simply copy over the restart files
            os.system(f'cp -L {fname_restart} {run_dir}')
        else:
            restart_options = self.perturb['restart']
            # copy restart files to the ensemble member directory
            fname = os.path.join(run_dir, os.path.basename(fname_restart))
            subprocess.run(['cp', fname_restart, fname])
            # prepare the restart file options for the perturbation
            file_options_rst = {'fname': fname,
                            'lon_name':file_options_restart['lon_name'],
                            'lat_name':file_options_restart['lat_name']}
            # perturb the restart file
            restart.perturb_restart(restart_options, file_options_rst, debug)

        if 'forcing' not in self.perturb:
            # we we do not perturb the forcing file
            # simply link the forcing files
            for forcing_name in file_options_forcing:
                os.system(f'ln -fs {fname_forcing[forcing_name]} {run_dir}')
        else:
            forcing_options = self.perturb['forcing']
            for forcing_name in forcing_options:
                # we ignore entries that are not in the files options
                # e.g., path
                if forcing_name not in fname_forcing: continue
                fname = os.path.join(run_dir,
                                    os.path.basename(fname_forcing[forcing_name])
                                    )
                # the forcing file options for the perturbation
                file_options_forcing[forcing_name]['fname_src'] = fname_forcing[forcing_name]
                file_options_forcing[forcing_name]['fname'] = fname
            # add forcing perturbations
            forcing.perturb_forcing(forcing_options, file_options_forcing, ens_mem_id, time, next_time, debug)

    def postprocess(self, task_id=0, **kwargs):
        """Postprocessing method for nextsim.dg
        Parameters: same as preprocess
        """
        kwargs = super().parse_kwargs(**kwargs)

        # get the current ensemble member id
        ens_mem_id:int = kwargs['member'] + 1
        # ensemble member directory for the current member
        ens_mem_dir:str = f'ens_{str(ens_mem_id).zfill(2)}'

        # directory where files are being collected to, and where the model will be run
        run_dir = os.path.join(kwargs['path'], ens_mem_dir)

        file = restart.get_restart_filename(self.files['restart'], ens_mem_id, kwargs['time'])
        restartfile = os.path.join(run_dir, os.path.basename(file))

        # read cice hice from restart file
        cice = nc_read_var(restartfile, 'data/cice')
        hice = nc_read_var(restartfile, 'data/hice')

        ##limit the cice between 0-1, and hice>0, before running the next forecast
        cice = dgLimit.limit_max(cice, 1.0)
        cice = dgLimit.limit_min(cice, 0.0)
        hice = dgLimit.limit_min(hice, 0.0)

        # write back to restart file
        dims = {'ydim':self.grid.ny, 'xdim':self.grid.nx, 'dg_comp':self.dg_comp}
        nc_write_var(restartfile, dims, 'data/cice', cice, comm=kwargs['comm'])
        nc_write_var(restartfile, dims, 'data/hice', hice, comm=kwargs['comm'])

    def run(self, task_id=0, **kwargs):
        """Run nextsim.dg model forecast"""
        kwargs = super().parse_kwargs(**kwargs)

        nproc = self.nproc_per_run
        offset = task_id*self.nproc_per_run

        time = kwargs['time']
        forecast_period = kwargs['forecast_period']
        next_time = time + forecast_period * dt1h

        member = kwargs['member']
        if member is not None:
            run_dir:str = os.path.join(kwargs['path'], f"ens_{member+1:02}")
        else:
            run_dir = kwargs['path']
        makedir(run_dir)
        namelist.make_namelist(self.files, self.model_config_file, run_dir, **kwargs)

        ##build shell commands for running the model
        shell_cmd = ""
        if self.model_env:
            shell_cmd += f". {self.model_env}; "
        shell_cmd += f"cd {run_dir}; "
        shell_cmd += "JOB_EXECUTE $NDG_BLD_DIR/nextsim --config-file nextsim.cfg > time.step"

        run_job(shell_cmd, job_name='nextsim.dg.run', parallel_mode=self.parallel_mode, nproc=nproc, offset=offset, run_dir=run_dir, **kwargs)

        ##check if the restart file at next_time is produced
        fname_restart = restart.get_restart_filename(self.files['restart'], 1, next_time)
        fname_out = os.path.join(run_dir, os.path.basename(fname_restart))
        if not os.path.exists(fname_out):
            raise RuntimeError(f"nextsim.dg.run: failed to produce {fname_out}, check {run_dir}")

    def run_batch(self, task_id=0, **kwargs):
        """Run nextsim.dg model ensemble forecast, use job array to spawn the member runs"""
        kwargs = super().parse_kwargs(**kwargs)
        assert self.use_job_array, \
            "use_job_array shall be True if running ensemble in batch mode."

        time = kwargs['time']
        forecast_period = kwargs['forecast_period']
        next_time = time + forecast_period * dt1h

        run_dir:str = kwargs['path']
        nens:int = kwargs['nens']
        for member in range(nens):
            ens_dir = os.path.join(run_dir, f"ens_{member+1:02}")
            makedir(ens_dir)

            kwargs['member'] = member
            ##this creates run_dir/ens_??/nextsim.cfg
            namelist.make_namelist(self.files, self.model_config_file,
                                   ens_dir, **kwargs)

        ##build shell commands for running the model using job array
        shell_cmd = "echo starting the script...; "
        shell_cmd += f". {self.model_env}; "
        shell_cmd += f"cd {run_dir}; "
        shell_cmd += f"echo {run_dir}; "
        shell_cmd += "cd ens_$(printf '%02d' JOB_ARRAY_INDEX); "
        shell_cmd += "echo $(pwd); "
        shell_cmd += "cp $NDG_BLD_DIR/nextsim .; "
        if self.parallel_mode == 'openmp':
            shell_cmd += f"export OMP_NUM_THREADS={self.nproc_per_run}; "
            shell_cmd += "./nextsim --config-file nextsim.cfg > time.step"
        elif self.parallel_mode == 'mpi':
            shell_cmd += "JOB_EXECUTE ./nextsim --config-file nextsim.cfg > time.step"
        else:
            raise TypeError(f"unknown parallel mode '{self.parallel_mode}'")

        run_job(shell_cmd, job_name='nextsim.dg.ens_run', use_job_array=self.use_job_array, nproc=self.nproc_per_run, walltime=self.walltime, array_size=nens, run_dir=run_dir, **kwargs)

        ##check if the restart files at next_time are produced
        fname_restart = restart.get_restart_filename(self.files['restart'], 1, next_time)
        for member in range(nens):
            ens_dir = os.path.join(run_dir, f"ens_{member+1:02}")
            fname_out = os.path.join(ens_dir, os.path.basename(fname_restart))
            if not os.path.exists(fname_out):
                raise RuntimeError(f"nextsim.dg.run_batch: failed to produce {fname_out}, check {ens_dir}")



import os
import numpy as np
from functools import lru_cache
from scipy.optimize import fsolve
from scipy.ndimage import distance_transform_edt, gaussian_filter
from NEDAS.utils.conversion import ensure_list
from NEDAS.utils.spatial_operation import gradx, grady, warp
from NEDAS.utils.fft_lib import fft2, ifft2, get_wn

def random_perturb(grid, fields, prev_perturb, dt=1, n=0, seed=None, **kwargs):
    """
    Add random perturbation to the given 2D field

    Args:
        grid: Grid object describing the 2d domain
        fields: list of np.array shape[...,ny,nx]
        prev_perturb; list of np.array from previous perturbation data, None if unavailable
        dt: float, interval (hours) between time steps
        n: int, current time step index
        variable: str, or list of str
        type: str: 'gaussian', 'powerlaw', or 'displace'
        amplitude: float, (or list of floats, in multiscale approach)
        hcorr: float, or list of float, horizontal corr length (meters)
        tcorr: float, or list of float, time corr length (hours)
        **kwargs: other arguments
    """
    if seed is None:
        ##try to randomize using system entropy
        seed = int.from_bytes(os.urandom(4), 'little')
    else:
        assert isinstance(seed, int)
    np.random.seed(seed)

    perturb = {}
    perturb_type, other_opts, params = parse_perturb_opts(**kwargs)

    for vname,rec in params.items():
        fld = fields[vname]
        assert grid.x.shape == fld.shape[-2:], f"input fields[{vname}] dimension mismatch with grid"

        ns = rec['nscale']
        if perturb_type == 'displace':
            perturb[vname] = np.zeros((ns,2)+fld.shape[-2:])
        else:
            perturb[vname] = np.zeros((ns,)+fld.shape)

        if prev_perturb[vname] is not None and n==0:
            perturb[vname] = prev_perturb[vname]
            continue

        ##loop over scale s and generate perturbation
        for s in range(ns):
            ##draw a random field for each 2d field component in fields
            for ind in np.ndindex(fld.shape[:-2]):
                if perturb_type == 'gaussian':
                    perturb[vname][(s,)+ind] = random_field_gaussian(grid.nx, grid.ny, rec['amp'][s], rec['hcorr'][s]/grid.dx)

                elif perturb_type == 'powerlaw':
                    perturb[vname][(s,)+ind] = random_field_powerlaw(grid.nx, grid.ny, rec['amp'][s], rec['powerlaw'][s])

                elif perturb_type == 'displace':
                    mask = np.full(grid.x.shape, False)
                    du, dv = random_displacement(grid, mask, rec['amp'][s], rec['hcorr'][s]/grid.dx)
                    perturb[vname][s] = np.array([du, dv])

                else:
                    raise NotImplementedError('unknown perturbation type: '+perturb_type)

            ##create perturbations that are correlated in time
            autocorr = 0.75
            ncorr = rec['tcorr'][s] / dt  ##time steps at decorrelation
            alpha = autocorr**(1.0 / ncorr)
            if prev_perturb[vname] is not None:
                perturb[vname][s] = np.sqrt(1-alpha**2) * perturb[vname][s] + alpha * prev_perturb[vname][s]

    ###legacy prsflg==1,2 options in force_perturb program, reproduced here
    if 'press_wind_relate' in other_opts:
        for vname in ['atmos_surf_velocity', 'atmos_surf_press']:
            assert vname in params.keys(), f'{vname} not in variable list, cannot run press_wind_relate option'

        for s in range(ns):
            perturb['atmos_surf_velocity'][s] = get_velocity_from_press(grid, perturb['atmos_surf_press'][s], ('scale_wind' in other_opts), params['atmos_surf_press']['amp'][s], params['atmos_surf_press']['hcorr'][s], params['atmos_surf_velocity']['amp'][s])

    ##now add perturbations to each field
    for vname,rec in params.items():
        for s in range(rec['nscale']):
            if perturb_type == 'displace':
                fields[vname] = warp(grid, fields[vname], perturb[vname][s,0,...], perturb[vname][s,1,...])

            else:
                if 'exp' in other_opts:
                    ##add lognormal perturbations
                    fields[vname] *= np.exp(perturb[vname][s,...] - 0.5*rec['amp'][s]**4)

                else:
                    ##just add the gaussian perturbations
                    fields[vname] += perturb[vname][s,...]

        ##respect value bounds after perturbing
        if 'bounds' in kwargs:
            vmin, vmax = kwargs['bounds']
            fields[vname] = np.minimum(np.maximum(fields[vname], vmin), vmax)

    return fields, perturb

def parse_perturb_opts(**kwargs):
    ##perturb['type'] string format:
    #main option (gaussian/powerlaw/displace) followed by , then additional options separated by ,
    opts = kwargs['type'].split(',')
    perturb_type = opts[0]
    other_opts = []
    for opt in opts[1:]:
        other_opts.append(opt)

    key_list = []
    for key in ['amp', 'hcorr', 'tcorr', 'powerlaw']:
        if key in kwargs:
            key_list.append(key)

    ##a list of variables can be specified if running a multivariate perturbation scheme
    ##form a variable list for further processing
    variable_list = ensure_list(kwargs['variable'])
    nv = len(variable_list)
    for key in key_list:
        kwargs[key] = ensure_list(kwargs[key])

    ##get perturbation parameters for each variable from kwargs
    params = {}
    for v in range(nv):
        vname = variable_list[v]
        params[vname] = {}
        ##in multiscale approach, a list of parameters can be specified for a variable;
        ##one separate perturbation will be generated for each, then they will be added together
        if isinstance(kwargs[key_list[0]][v], list):
            nscale = len(kwargs[key_list[0]][v])
        else:
            nscale = 1
            for key in key_list:  ##make a list even if only one value for the key
                kwargs[key][v] = [kwargs[key][v]]
        params[vname]['nscale'] = nscale
        ##check if all keys are lists with same len
        for key in key_list[1:]:
            assert len(kwargs[key][v]) == nscale, f"perturb option: {key} has different number of entries from {key_list[0]}, check config"
        ##assign the parameters
        for key in key_list:
            params[vname][key] = kwargs[key][v]

    return perturb_type, other_opts, params

def get_velocity_from_press(grid, pres, scale_wind=False, press_amp=None, press_hcorr=None, wind_amp=None):
    """derive wind velocity from pressure field """
    rhoa = 1.2
    r2d = 180/np.pi
    wlat = 15.
    plon, plat = grid.proj(grid.x, grid.y, inverse=True)
    fcor = 2 * np.sin(40./r2d)* 2*np.pi / 86400  ##coriolis at 40N

    ##grid spacing
    dx = grid.dx / grid.mfx
    dy = grid.dy / grid.mfy
    ##hcorr = rh * dx according to mid-domain dx (to be consistent with ReanalysisTP5/Perturb_forcing)
    dx_ = grid.dx / grid.mfx[grid.ny//2, grid.nx//2]

    wprsfac = 1.
    if scale_wind:
        ds = 0.54 * press_hcorr / dx_ * np.hypot(dx, dy)  ##horizontal scale
        wind_scale = press_amp / ds / fcor / rhoa  ##expected wind scale from pressure field
        wprsfac = wind_amp / wind_scale  ##scaling factor to match wind perturbation with given amp

    ##pres gradients
    dpresx = gradx(pres, dx, grid.cyclic_dim) * wprsfac
    dpresy = grady(pres, dy, grid.cyclic_dim) * wprsfac

    ##geostrophic wind near poles
    vcor =  dpresx / fcor / rhoa * np.sign(plat)
    ucor = -dpresy / fcor / rhoa * np.sign(plat)

    ##gradient wind near equator
    ueq = -dpresx / fcor / rhoa
    veq = -dpresy / fcor / rhoa

    wcor = np.sin(np.minimum(wlat, plat) / wlat * np.pi * 0.5)
    u = wcor*ucor + (1-wcor)*ueq
    v = wcor*vcor + (1-wcor)*veq

    return np.array([u, v])

def random_field_gaussian(nx, ny, amp, hcorr):
    """
    Random field with a Gaussian spectrum.

    Args:
        nx (int): Number of grid points in x direction.
        ny (int): Number of grid points in y direction.
        amp (float): Amplitude (standard deviation) of the random field.
        hcorr (float): Horizontal decorrelation length (number of grid points).

    Returns:
        np.ndarray: The output random field with shape (ny, nx).
    """
    fld = np.zeros((ny, nx))
    kx, ky = get_wn(fld)
    k2d = np.hypot(kx, ky)
    sig_out = get_sig_in_gaussian(nx, ny, hcorr)

    ##draw random phase from a white noise field
    ph = fft2(np.random.normal(0, 1, fld.shape))

    ##scaling factor to get the right amplitude
    norm2 = np.sum(gaussian(k2d, sig_out)**2) / (nx*ny)
    sf = amp / np.sqrt(norm2)
    return np.real(ifft2(sf * gaussian(k2d, sig_out) * ph))

def gaussian(k, sig):
    """gaussian spectrum"""
    return np.exp(- k**2 / sig**2)

@lru_cache
def get_sig_in_gaussian(nx, ny, hcorr):
    """
    Derive the sig parameter in gaussian spectrum.
    Called by random_field_gaussian, cached the result since in external for loop there will be fixed input
    """
    fld = np.zeros((ny, nx))
    nup = int(max(nx, ny))
    kx, ky = get_wn(fld)
    k2d = np.hypot(kx, ky)

    def func2d(sig):
        sum1 = np.sum(gaussian(k2d, sig)**2 * np.cos(2*np.pi * kx/nup * hcorr))
        sum2 = np.sum(gaussian(k2d, sig)**2)
        return sum1/sum2 - np.exp(-1)

    ##solve for sig given hcorr so that func2d=0
    ##first deal with some edge cases
    if hcorr <= 2:  ##shorter than resolved scale, sig_out should be largest
        sig_out = nup
    elif hcorr >= nup/2:  ##longer than domain scale, sig_out should be smallest
        sig_out = 1.
    else:
        sig_out = np.abs(fsolve(func2d, 1)[0])
    # print(sig_out, func2d(sig_out))
    return sig_out

def random_field_powerlaw(nx, ny, amp, pwrlaw):
    fld = np.zeros((ny, nx))
    kx, ky = get_wn(fld)
    k2d = np.hypot(kx, ky)

    k2d[np.where(k2d==0.0)] = 1e-10  ##avoid singularity, set wn-0 to small value

    ##draw random phase from a white noise field
    ph = fft2(np.random.normal(0, 1, fld.shape))

    ##assemble random field given amplitude from power spectrum, and random phase ph
    norm = 2 * np.pi * k2d
    amp_ = amp * np.sqrt(k2d**(pwrlaw+1) / norm)
    amp_[np.where(k2d==1e-10)] = 0.0 ##zero amp for wn-0
    return np.real(ifft2(amp_ * ph))

def random_displacement(grid, mask, amp, hcorr):
    ##prototype: generate a random wavenumber-1 displacement for the whole domain

    ##set the boundaries to be masked if they are not cyclic
    ##i.e. we don't want displacement across non-cyclic boundary
    if grid.cyclic_dim is not None:
        cyclic_dim = grid.cyclic_dim
    else:
        cyclic_dim = ''
    if 'x' not in cyclic_dim:
        mask[:, 0] = True
        mask[:, -1] = True
    if 'y' not in cyclic_dim:
        mask[0, :] = True
        mask[-1, :] = True

    ##distance to masked area
    dist = distance_transform_edt(1-mask.astype(float))
    dist /= np.max(dist)
    dist = gaussian_filter(dist, sigma=10)
    dist[mask] = 0

    ##the displacement vector field from a random streamfunction
    psi = random_field_gaussian(grid.nx, grid.ny, 1, hcorr)

    du, dv = -grady(psi, 1, grid.cyclic_dim), gradx(psi, 1, grid.cyclic_dim)
    norm = np.hypot(du, dv).max()
    du *= amp / norm
    dv *= amp / norm

    du *= dist  ##taper near mask
    dv *= dist
    return du, dv




from . import assimilators, covariance, inflation, localization, state, obs, transforms, updators

__all__ = ['assimilators', 'covariance', 'inflation', 'localization', 'state', 'obs', 'transforms', 'updators']


from NEDAS.config import Config

registry = {
    'gaspari_cohn': ('distance_based', 'gaspari_cohn_func'),
    'step': ('distance_based', 'step_func'),
    'exponential': ('distance_based', 'exponential_func'),
}

def get_localization_funcs(c: Config) -> dict:
    local_funcs = {}
    for key in ['horizontal', 'vertical', 'temporal']:
        if c.localization_def[key]:
            if 'type' not in c.localization_def[key]:
                raise KeyError(f"'type' needs to be specified for c.localization['{key}']")
            local_funcs[key] = get_localization_func_component(c.localization_def[key]['type'])
        else:
            local_funcs[key] = None
    return local_funcs

def get_localization_func_component(localization_types):
    localization_types = localization_types.lower().split(',')

    ##distance-based localization schemes
    if 'gaspari_cohn' in localization_types:
        from .distance_based import gaspari_cohn_func as local_func
    elif 'step' in localization_types:
        from .distance_based import step_func as local_func
    elif 'exponential' in localization_types:
        from .distance_based import exponential_func as local_func

    # ##correlation based localization schemes
    # elif 'SER' in localization_types:
    #     from NEDAS.assim_tools.localization.SER import CorrelationBasedLocalization as Localization
    # elif 'NICE' in localization_types:
    #     from NEDAS.assim_tools.localization.NICE import 
    else:
        raise ValueError(f"Unknown localization type {type}")
    return local_func



import os
import subprocess
import numpy as np
from NEDAS.utils.conversion import t2s, dt1h
from ..util import read_data_bin, write_data_bin, grid2spec, spec2grid
from ..model import QGModel

class QGModelEmulator(QGModel):
    """
    Class for configuring and running the qg model emulator
    """
    def run_batch(self, nens=1, task_id=0, **kwargs):
        kwargs = super().super().parse_kwargs(**kwargs)
        self.run_status = 'running'

        ##load model weights if not yet
        if self.unet_model is None:
            from .netutils import Att_Res_UNet
            self.unet_model = Att_Res_UNet(**self.model_params).make_unet_model()
            self.unet_model.load_weights(self.weightfile)

        if nens>1:
            ##running ensmeble together, ignore kwargs['member']
            members = list(range(nens))
        else:
            members = [kwargs['member']]

        path = kwargs['path']
        time = kwargs['time']
        forecast_period = kwargs['forecast_period']
        nsteps = int(forecast_period / self.restart_dt)
        next_time = time + forecast_period * dt1h

        ##input ensemble state
        state = np.zeros((nens,self.ny,self.nx,self.nz))
        for m in range(nens):
            kwargs_in = {**kwargs, 'member':members[m], 'time':time}
            input_file = self.filename(**kwargs_in)
            run_dir = os.path.dirname(input_file)
            subprocess.run("mkdir -p "+run_dir, shell=True)

            input_file = self.filename(**kwargs_in)
            for k in range(self.nz):
                psik = read_data_bin(input_file, self.kmax, self.nz, k)
                state[m,...,k] = spec2grid(psik).T

        ##run prediction model
        for i in range(nsteps):
            state = self.unet_model.predict(state, verbose=1)
        state_out = state

        ##output to restart file
        for m in range(nens):
            kwargs_out = {**kwargs, 'member':members[m], 'time':next_time}
            output_file = self.filename(**kwargs_out)
            if not os.path.exists(output_file):
                with open(output_file, 'wb'):
                    pass
            for k in range(self.nz):
                fld = state_out[m,...,k]
                self.write_var(fld, name='streamfunc', k=k, **kwargs_out)

    def run(self, task_id=0, **kwargs):
        raise NotImplementedError("qg.emulator only runs in batch mode, call run_batch instead")



import numpy as np
import os
import glob
from datetime import datetime, timedelta, timezone
import netCDF4
import pyproj
from NEDAS.grid import Grid
from NEDAS.datasets import Dataset

class OsisafSeaIceDriftObs(Dataset):

    def __init__(self, config_file=None, parse_args=False, **kwargs):
        super().__init__(config_file, parse_args, **kwargs)

        self.variables = {'seaice_drift': {'dtype':'float', 'is_vector':True, 'z_units':'m', 'units':'km/day'}, }

        proj = pyproj.Proj(self.proj)
        x, y = np.meshgrid(np.arange(self.xstart, self.xend, self.dx), np.arange(self.ystart, self.yend, self.dy))
        self.grid = Grid(proj, x, y)

        self.obs_operator = {'seaice_drift': self.get_seaice_drift,}

    def filename(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        path = kwargs['path']
        time = kwargs['time']
        name = kwargs['name']
        obs_window_min = kwargs['obs_window_min']
        obs_window_max = kwargs['obs_window_max']

        if time is None:
            search = os.path.join(path, '????', '??', 'ice_drift_'+self.proj_name+'-625_multi-oi_????????????-????????????.nc')
            file_list = glob.glob(search)

        else:
            if obs_window_min is not None and obs_window_max is not None:
                d_range = np.arange(obs_window_min, obs_window_max)
            else:
                d_range = [0]

            file_list = []
            for d in d_range:
                t = time + d * timedelta(hours=1)
                t0 = t - timedelta(days=2)  ##drift traj start time
                search = os.path.join(path, t.strftime('%Y'), t.strftime('%m'), 'ice_drift_'+self.proj_name+'-625_multi-oi_'+t0.strftime('%Y%m%d%H%M')+'-'+t.strftime('%Y%m%d%H%M')+'.nc')
                for result in glob.glob(search):
                    if result not in file_list:
                        file_list.append(result)

        assert len(file_list)>0, 'no matching files found: '+search
        return file_list

    def read_obs(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        grid = kwargs['grid']
        mask = kwargs['mask']

        self.grid.set_destination_grid(grid)  ##for vector rotation

        obs_seq = {'obs':[], 't':[], 'z':[], 'y':[], 'x':[], 'err_std':[], }

        for fname in self.filename(**kwargs):

            ##read the data file
            f = netCDF4.Dataset(fname)

            lat = f['lat'][...].data.flatten()
            lon = f['lon'][...].data.flatten()
            x_, y_ = grid.proj(lon, lat)
            mask_ = grid.interp(mask.astype(int), x_, y_)

            ntime = f.dimensions['time'].size
            for n in range(ntime):
                t0, t1 = f['time_bnds'][n,:].data
                obs_dt0 = f['dt0'][n,...].data.flatten()
                obs_dt1 = f['dt1'][n,...].data.flatten()
                qc_flag = f['status_flag'][n,...].data.flatten()

                ###get drift vector, rotate vector from proj to grid.proj
                obs_dx = f['dX'][n,...]
                obs_dy = f['dY'][n,...]
                obs_drift = self.grid.rotate_vectors(np.array([obs_dx, obs_dy]))
                obs_dx = obs_drift[0,...].flatten()
                obs_dy = obs_drift[1,...].flatten()

                if 'uncert_dX_and_dY' in f.variables:
                    obs_err = f['uncert_dX_and_dY'][n,...].data.flatten()
                else:
                    obs_err = 10 * np.ones(obs_dx.shape)  ##default drift err 10 km

                for p in range(obs_dx.size):
                    if qc_flag[p] != 30:
                        continue
                    if obs_err[p] <= 0:
                        continue

                    if x_[p] < grid.xmin or x_[p] > grid.xmax or y_[p] < grid.ymin or y_[p] > grid.ymax:
                        continue

                    if mask_[p] > 0:
                        continue

                    dt0 = obs_dt0[p] if ~np.isnan(obs_dt0[p]) else 0.
                    dt1 = obs_dt1[p] if ~np.isnan(obs_dt1[p]) else 0.
                    obs_t = (t1+dt1) * timedelta(seconds=1) + datetime(1978, 1, 1, tzinfo=timezone.utc)
                    obs_dt = (t1+dt1 - t0-dt0)*timedelta(seconds=1) / timedelta(days=1)
                    obs_u = obs_dx[p] / obs_dt
                    obs_v = obs_dy[p] / obs_dt
                    obs_value = [obs_u, obs_v]

                    obs_err_std = obs_err[p] / obs_dt  ##uncertainty from dataset

                    ##assignn to obs_seq
                    obs_seq['obs'].append(obs_value)
                    obs_seq['err_std'].append(obs_err_std)
                    obs_seq['t'].append(obs_t)
                    obs_seq['z'].append(0)
                    obs_seq['y'].append(y_[p])
                    obs_seq['x'].append(x_[p])

            f.close()

        for key in obs_seq.keys():
            obs_seq[key] = np.array(obs_seq[key])

        ##make obs dimension [2,nobs] for vectors
        obs_seq['obs'] = obs_seq['obs'].T

        return obs_seq

    def get_seaice_drift(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        grid = kwargs['grid']
        path = kwargs['path']
        member = kwargs['member']
        model = kwargs['model']
        model.grid.set_destination_grid(grid)

        start_x = kwargs['x']
        start_y = kwargs['y']
        start_t = kwargs['t']
        x = start_x.copy()
        y = start_y.copy()
        obs_seq = np.zeros((2,)+x.shape)

        return obs_seq



import numpy as np
from NEDAS.grid import Grid
from NEDAS.utils.multiscale import get_scale_component, get_error_scale_factor
from .base import Transform

class ScaleBandpass(Transform):
    """
    Subclass for scale bandpass filter to get a scale component.
    """
    def __init__(self, decompose_obs=True, **kwargs):
        self.decompose_obs = decompose_obs

    def forward_state(self, c, rec, field):
        if c.nscale == 1:
            return field

        ##for state on analysis grid, just get_scale_component (using Fourier method)
        ##pad voids with zero
        mask = np.isnan(field)
        field[mask] = 0.0
        field = get_scale_component(c.grid, field, c.character_length, c.iter)
        field[mask] = np.nan
        return field

    def backward_state(self, c, rec, field):
        return field

    def forward_obs(self, c, obs_rec, obs_seq):
        if c.nscale == 1:
            return obs_seq

        if not self.decompose_obs:
            return obs_seq

        ##temporarily convert obs grid to the analysis grid
        ##create the irregular obs grid
        obs_grid = Grid(c.grid.proj, obs_seq['x'], obs_seq['y'], regular=False)

        ##remove unwanted triangles in the obs grid
        max_a = np.quantile(obs_grid.tri.a, 0.999)
        max_p = np.quantile(obs_grid.tri.p, 0.99)
        msk = np.logical_or(obs_grid.tri.a > max_a, obs_grid.tri.p > max_p, obs_grid.tri.ratio < 0.3)

        ##convert obs to analysis grid
        obs_grid = Grid(c.grid.proj, obs_seq['x'], obs_seq['y'], regular=False, triangles=obs_grid.tri.triangles[~msk,:])
        obs_grid.set_destination_grid(c.grid)
        obs_fld = obs_grid.convert(obs_seq['obs'], is_vector=obs_rec['is_vector'], method='nearest', coarse_grain=False)

        ##pad voids with zeros, for convolution later
        mask = np.isnan(obs_fld)
        obs_fld[mask] = 0.0

        ##get scale component on analysis grid
        obs_fld_new = get_scale_component(c.grid, obs_fld, c.character_length, c.iter)
        if obs_rec['is_vector']:
            for i in range(2):
                obs_seq['obs'][i,...] = c.grid.interp(obs_fld_new[i,...], obs_seq['x'], obs_seq['y'], method='nearest')
        else:
            obs_seq['obs'] = c.grid.interp(obs_fld_new, obs_seq['x'], obs_seq['y'], method='nearest')

        ##TODO: current implementation is very slow
        ##update obs err std because some averaging happened in get_scale_component
        #obs_seq['err_std'] *= get_error_scale_factor(c.grid, c.character_length, c.iter)
        obs_seq['err_std'] *= c.obs_err_scale_fac[c.iter]

        return obs_seq

    def backward_obs(self, c, obs_rec, obs_seq):
        return obs_seq



##check if your mpi environment is correctly setup
import numpy as np
import unittest
from NEDAS.utils.parallel import Comm, distribute_tasks

class TestParallel(unittest.TestCase):

    def test_parallel_start(self):
        comm = Comm()
        self.assertIsInstance(comm.Get_size(), int)
        self.assertIsInstance(comm.Get_rank(), int)
        self.assertTrue(comm.Get_rank() < comm.Get_size())


    def test_bcast(self):
        comm = Comm()

        if comm.Get_rank() == 0:
            data0 = 100
            data1 = np.array([1, 2, 3])
            data2 = {'x':1, 'y':2}
            data3 = [1, True]
        else:
            data0 = None
            data1 = None
            data2 = None
            data3 = None
        data0 = comm.bcast(data0)
        data1 = comm.bcast(data1)
        data2 = comm.bcast(data2)
        data3 = comm.bcast(data3)

        self.assertEqual(data0, 100)
        self.assertTrue((data1==np.array([1,2,3])).all())
        self.assertEqual(data2['x'], 1)
        self.assertEqual(data2['y'], 2)
        self.assertEqual(data3[0], 1)
        self.assertTrue(data3[1])


    def test_send_recv(self):
        comm = Comm()
        pid = comm.Get_rank()
        nproc = comm.Get_size()
        if pid == 0:
            data = np.array([1, 2, 3])
            comm.send(data, dest=nproc-1, tag=0)
        if pid == nproc-1:
            recv_data = comm.recv(source=0, tag=0)
            self.assertTrue((recv_data == np.array([1, 2, 3])).all())


    def test_allgather(self):
        comm = Comm()
        pid = comm.Get_rank()
        nproc = comm.Get_size()
        data = [pid]

        gather_data = []
        for entry in comm.allgather(data):
            for value in entry:
                gather_data.append(value)

        if pid == 0:
            self.assertTrue((np.array(gather_data) == np.arange(nproc)).all())


    def test_distribute_tasks(self):
        comm = Comm()
        pid = comm.Get_rank()
        nproc = comm.Get_size()

        ##case1: one task per pid
        task_list = distribute_tasks(comm, np.arange(nproc))
        self.assertEqual(task_list[pid][0], pid)

        ##case2: more tasks than nproc
        ntasks = 10*nproc
        task_list = distribute_tasks(comm, np.arange(ntasks))
        full_task_list = []
        for p in range(nproc):
            for e in task_list[p]:
                full_task_list.append(e)
        self.assertTrue((np.array(full_task_list) == np.arange(ntasks)).all())

        ##case3: fewer tasks than nproc
        ntasks = np.max(1, nproc//10)
        task_list = distribute_tasks(comm, np.arange(ntasks))
        full_task_list = []
        for p in range(nproc):
            for e in task_list[p]:
                full_task_list.append(e)
        self.assertTrue((np.array(full_task_list) == np.arange(ntasks)).all())

        ##case4: uneven workload for 100 tasks
        ntasks = 100
        workload = np.random.randint(1, 10, ntasks)
        task_list = distribute_tasks(comm, np.arange(ntasks), workload)
        full_task_list = []
        for p in range(nproc):
            for e in task_list[p]:
                full_task_list.append(e)
        self.assertTrue((np.array(full_task_list) == np.arange(ntasks)).all())


if __name__ == '__main__':
    unittest.main()




import os
import inspect
from abc import ABC, abstractmethod
import numpy as np
from NEDAS.config import parse_config
from NEDAS.utils.parallel import bcast_by_root
from NEDAS.utils.progress import timer

class Assimilator(ABC):
    def __init__(self, c):
        self.analysis_dir = c.analysis_dir(c.time, c.iter)

        ##get parameters from config file
        code_dir = os.path.dirname(inspect.getfile(self.__class__))
        config_dict = parse_config(code_dir, parse_args=False, **c.assimilator_def)
        for key, value in config_dict.items():
            setattr(self, key, value)

    def assimilate(self, c, state, obs):
        """
        Main method to run the batch assimilation algorithm
        """
        self.prior_inflation(c, state, obs)
        self.partition_grid(c, state, obs)
        timer(c)(self.transpose_to_ensemble_complete)(c, state, obs)
        timer(c)(self.assimilation_algorithm)(c, state, obs)
        self.transpose_to_field_complete(c, state, obs)
        self.posterior_inflation(c, state, obs)

    def prior_inflation(self, c, state, obs):
        """
        Apply covariance inflation for the prior ensemble
        """
        state.output_ens_mean(c, state.fields_prior, state.prior_mean_file)
        c.inflation_func(c, state, obs, 'prior')
        state.output_state(c, state.fields_prior, state.prior_file)

    def posterior_inflation(self, c, state, obs):
        """
        Apply covariance inflation for the posterior ensemble
        """
        obs.prepare_obs_from_state(c, state, 'posterior')  ##update obs_post_seq for stats
        state.output_ens_mean(c, state.fields_post, state.post_mean_file)
        c.inflation_func(c, state, obs, 'posterior')
        state.output_state(c, state.fields_post, state.post_file)

    def partition_grid(self, c, state, obs):
        """
        Partition the analysis grid into several parts and distribute the workload over the mpi ranks.
        """
        state.partitions = bcast_by_root(c.comm)(self.init_partitions)(c)
        obs.obs_inds = bcast_by_root(c.comm_mem)(self.assign_obs)(c, state, obs)
        state.par_list = bcast_by_root(c.comm)(self.distribute_partitions)(c, state, obs)

    def init_partitions(self, c):
        raise NotImplementedError

    def assign_obs(self, c, state, obs):
        raise NotImplementedError

    def distribute_partitions(self, c, state, obs):
        raise NotImplementedError

    def transpose_to_ensemble_complete(self, c, state, obs):
        """
        Communicate among mpi ranks and transpose the locally-stored state/obs chunks to ensemble-complete
        """
        c.print_1p('>>> transpose to ensemble complete:\n')

        c.print_1p('state variables: ')
        state.state_prior = state.transpose_to_ensemble_complete(c, state.fields_prior)

        c.print_1p('z coords: ')
        state.z_state = state.transpose_to_ensemble_complete(c, state.z_fields)

        c.print_1p('obs sequences: ')
        obs.lobs = obs.transpose_to_ensemble_complete(c, state, obs.obs_seq)

        c.print_1p('obs prior sequences: ')
        obs.lobs_prior = obs.transpose_to_ensemble_complete(c, state, obs.obs_prior_seq, ensemble=True)

        if c.debug:
            np.save(os.path.join(self.analysis_dir, f'state_prior.{c.pid_mem}.{c.pid_rec}.npy'), state.state_prior)
            np.save(os.path.join(self.analysis_dir, f'z_state.{c.pid_mem}.{c.pid_rec}.npy'), state.z_state)
            np.save(os.path.join(self.analysis_dir, f'lobs.{c.pid_mem}.{c.pid_rec}.npy'), obs.lobs)
            np.save(os.path.join(self.analysis_dir, f'lobs_prior.{c.pid_mem}.{c.pid_rec}.npy'), obs.lobs_prior)

    def transpose_to_field_complete(self, c, state, obs):
        """
        Communicate among mpi ranks and transpose the locally-stored state/obs chunks
        back to field-complete
        """
        c.print_1p('>>> transpose back to field complete\n')

        c.print_1p('state variables: ')
        state.fields_post = state.transpose_to_field_complete(c, state.state_post)

        # c.print_1p('obs prior sequences: ')
        ##TODO there is a bug here, in transpose seq[:, ind] out of bound
        # obs.obs_post_seq = obs.transpose_to_field_complete(c, state, obs.lobs_post)

    @abstractmethod
    def assimilation_algorithm(self, c, state, obs) -> None:
        """
        The main assimilation algorithm will be implemented by subclasses
        """
        pass


import subprocess
from .slurm import SLURMJobSubmitter

class BetzyJobSubmitter(SLURMJobSubmitter):
    """
    JobSubmitter subclass for Norwegian betzy.sigma2.no supercomputers
    """

    def check_resources(self):
        super().check_resources()

        ##don't allow more than 6 processors on betzy login node
        p = subprocess.run("hostname", capture_output=True, text=True)
        if p.stdout.strip()[:5] == 'login':
            assert self.nproc+self.offset < 6, "Unsafe to run more than 6 processors on Betzy login node, aborting"



###util func for diagnostics
import numpy as np
from NEDAS.utils.fft_lib import fft2, get_wn

def rmse(fld, fld_tr):
    return

def pattern_corr(fld, fld_tr):
    return

##some spectral diagnostics
def pwrspec2d(fld):
    """
    Horizontal 2D power spectrum p(k2d), on a regular Cartesian grid

    For very large grid, the map factors will cause slight errors in grid spacings
    but okay for relatively small grids. Of course for global analysis one shall use
    the spherical harmonics instead.

    Input:
    - fld: np.array, [..., ny, nx]
      n-dimensional input field, the last two dimensions are the horizontal directions (y,x)

    Returns:
    - wn: np.array, [nup]
      Wavenumber in 2D, int(k2d), nup is the max wavenumber given ny,nx (whichever is larger)

    - pwr: np.array, [..., nup]
      The power spectrum, leading dimensions the same as fld but ny,nx replaced by nup.
    """
    ny, nx = fld.shape[-2:]
    kx, ky = get_wn(fld)
    nup = int(max(kx.max(), ky.max()))

    ##2d total wavenumber
    k2d = np.hypot(kx, ky)

    ##2d fft of fld, convert to power (variance)
    P = (np.abs(fft2(fld))/ny/nx)**2

    ##sum all kx,ky points with same k2d
    wn = np.arange(0., nup)
    pwr = np.zeros(fld.shape[:-2] + (nup,))
    for k in range(nup):
        pwr[..., k] = np.mean(P[np.where(np.ceil(k2d)==k)])
        ##we show mean pwr spectrum, it will be more intuitive to
        ##see the white noise spectrum as a flat line;
        ##
        ##however, in turbulence or atmospheric sciences, it is typical to
        ##show pwr = sum(P[k2d==w]); where the white noise has +1 slope
        ##and typical synoptic scale flows have a well-known -3 slope

    return wn, pwr

###some ensemble metrics
def crps(fld_ens, fld_tr):
    return




from datetime import datetime, timedelta, timezone

def diy(iyr):
   if ( (iyr%4 == 0 and 1901+iyr%400==0) or (iyr % 4 != 0 )):
      days=365
   else:
      days=366
   return days

def forday(yrflag,dtime):
   ## converts model day to "calendar" date (year,ordinal-day,hour).
   if (yrflag==0):
      #  360 days per model year, starting Jan 16
      iyear =  int((dtime+15.001)/360) + 1
      iday  =  (dtime+15.001) % 360. + 1
      ihour = ( ( dtime+15.001) % 360. + 1. - iday)*24.
   elif (yrflag==1):
      # 366 days per model year, starting Jan 16
      iyear =  int((dtime+15.001)/366.) + 1
      iday  =  ( dtime+15.001 ) % 366. + 1
      ihour =  ( ( dtime+15.001 ) % 366. + 1. - iday)*24.
   elif (yrflag==2):
      # 366 days per model year, starting Jan 01
      iyear =  int((dtime+ 0.001)/366.) + 1.
      iday  =  ( dtime+ 0.001 ) % 366. + 1.
      ihour = ( ( dtime+ 0.001 ) % 366. + 1. - iday)*24.
   elif (yrflag==3):
      # model day is calendar days since 01/01/1901
      iyr   = int((dtime-1.)/365.25)
      nleap = iyr/4
      dtim1 = 365.*iyr + nleap + 1.
      day   = dtime - dtim1 + 1.
      if     (dtim1>dtime):
         iyr = iyr - 1
      elif (day>=367.):
         iyr = iyr + 1
      elif ( (day>=366.) and (iyr % 4 != 3) ):
         iyr = iyr + 1
      nleap = int(iyr/4)
      dtim1 = 365.*iyr + nleap + 1.
      iyear =  1901 + iyr
      iday  =  dtime - dtim1 + 1.001
      ihour = (dtime - dtim1 + 1.001 - iday)*24
   else:
      print("unknown flag")
   return  int(iyear) , int(iday), int(ihour)

def dayfor(yrflag, iyear,iday,ihour):
   ##converts "calendar" date (year,ordinal-day,hour) to model day
   if     (yrflag==0):
      # 360 days per model year, starting Jan 16
      dtime =  (iyear-1) * 360. +  iday  + ihour/24. - 16.
   elif (yrflag==1):
      # 366 days per model year, starting Jan 16
        dtime = (iyear-1) * 366. + iday  + ihour/24.- 16.
   elif (yrflag==2):
      # 366 days per model year, starting Jan 01
      dtime = (iyear-1) * 366. + iday   - ihour/24. - 1.
   elif (yrflag==3):
      #model day is calendar days since 01/01/1901
      iyr=iyear-1901
      dtime=0.
      for k in range (iyr):
         dtime=dtime+diy(k+1901)

      iday2=iday
      while (iday2+ihour/24.>0.):
         iyr=iyr+1
         diy2=diy(iyr+1901)

         if ((iday2+(ihour/24.))/diy2>=1.):
            dtime=dtime+diy2
         else:
            dtime=dtime+iday2+ihour/24.
         iday2=iday2-diy2
   return dtime

def datetojul(t, refyear=1950, refmonth=1, refday=1):
    ref_t = datetime(refyear, refmonth, refday, tzinfo=timezone.utc)
    jday = (t - ref_t) / timedelta(days=1)
    return jday

def jultodate(jday, refyear=1950, refmonth=1, refday=1):
    ref_t = datetime(refyear, refmonth, refday, tzinfo=timezone.utc)
    t = ref_t + jday * timedelta(days=1)
    return t




import numpy as np
import struct
from NEDAS.utils.conversion import type_convert, type_dic, type_size

def get_info(filename):
    """basic read/write of data from nextsim bin/dat file"""
    info = {}
    with open(filename.replace('.bin','.dat'), 'r') as f:
        lines = f.readlines()

        pos = 4
        for recno, lin in enumerate(lines):
            ss = lin.split()
            if len(ss) < 2:
                continue
            ###get info about the variable record
            v_name = ss[0]
            v_type = ss[1]
            v_len = int(ss[2])
            typeconv = type_convert[v_type]
            v_min_val = typeconv(ss[3])
            v_max_val = typeconv(ss[4])
            v_rec = {}
            v_rec.update({'type':v_type})
            v_rec.update({'len':v_len})
            v_rec.update({'min_val':v_min_val})
            v_rec.update({'max_val':v_max_val})
            v_rec.update({'recno':recno})
            v_rec.update({'pos':pos})

            ##add variable record to the info dict
            info.update({v_name: v_rec})

            ##position of starting point for f.seek
            pos += (4 + v_len *type_size[v_type])
    return info

def read_data(filename, v_name):
    """Read data frim a binary file"""
    info = get_info(filename)
    if v_name not in info:
        raise ValueError('variable %s not found in %s' %(v_name, filename))

    with open(filename, 'rb') as f:
        f.seek(info[v_name]['pos'])
        v_len = info[v_name]['len']
        v_type = info[v_name]['type']
        v_data = np.array(struct.unpack((v_len*type_dic[v_type]),
                                        f.read(v_len*type_size[v_type])))
    return v_data

def write_data(filename, v_name, v_data):
    """
    write some var to existing nextsim restart file
    only for outputing analysis after DA
    to generate new file, see nextsim model documentation
    """
    info = get_info(filename)
    if v_name not in info:
        raise ValueError('variable %s not in %s' %(v_name, filename))
    ##update info with new v_data
    assert v_data.dtype == type_convert[info[v_name]['type']], "input data type mismatch with datfile record"
    assert v_data.size == info[v_name]['len'], f"input data size {v_data.size} mismatch with datfile record {info[v_name]['len']}"
    info[v_name]['min_val'] = np.nanmin(v_data)
    info[v_name]['max_val'] = np.nanmax(v_data)

    ##write the dat file with updated info
    with open(filename.replace('.bin','.dat'), 'wt') as f:
        for v in info:
            ss = '%s %s %i' %(v, info[v]['type'], info[v]['len'])
            if info[v]['type'] == 'int':
                ss += ' %i %i' %(info[v]['min_val'], info[v]['max_val'])
            else:
                ss += ' %g %g' %(info[v]['min_val'], info[v]['max_val'])
            f.write(ss+'\n')

    ##write v_data to the bin file
    with open(filename, 'r+b') as f:
        if len(v_data) != info[v_name]['len']:
            raise ValueError('v_data length does not match length on record')
        f.seek(info[v_name]['pos'])
        f.write(v_data.tobytes())



"""Diagnostic module to plot the ensemble states"""

import os
import numpy as np
import matplotlib.pyplot as plt
from NEDAS.utils.conversion import ensure_list, dt1h
from NEDAS.utils.shell_utils import makedir
from NEDAS.utils.graphics import add_colorbar, adjust_ax_size, get_cmap
from NEDAS.assim_tools.state import State

def get_task_list(c, **kwargs) -> list:

    variables = ensure_list(kwargs['variables'])
    vmin_diff = ensure_list(kwargs['vmin_diff'])
    vmax_diff = ensure_list(kwargs['vmax_diff'])
    nlevels_diff = ensure_list(kwargs['nlevels_diff'])
    cmap_diff = ensure_list(kwargs['cmap_diff'])

    state = State(c)

    tasks = []
    for member in range(c.nens):
        for i, vname in enumerate(variables):
            levels = [r['k'] for id,r in state.info['fields'].items() if r['name']==vname]
            assert len(levels)>0, f"cannot find state variable '{vname}'"
            for k in levels:
                for t in c.time + np.array(c.state_time_steps) * dt1h:
                    tasks.append({**kwargs, 'time':t, 'member':member, 'vname':vname, 'k':k, 'vmin_diff':vmin_diff[i], 'vmax_diff':vmax_diff[i], 'nlevels_diff':nlevels_diff[i], 'cmap_diff':cmap_diff[i]})
    return tasks

def run(c, **kwargs) -> None:
    """
    Run diagnostics: plot the ensemble states
    """
    if 'plot_dir' in kwargs:
        plot_dir = kwargs['plot_dir']
    else:
        plot_dir = os.path.join(c.work_dir, 'plots', 'analysis_increments')
    makedir(plot_dir)

    figsize = (kwargs.get('fig_size_x', 9), kwargs.get('fig_size_y', 8))
    landcolor = kwargs.get('land_color', 'gray')

    variables = ensure_list(kwargs['variables'])
    vname = kwargs['vname']
    vmin_diff = kwargs['vmin_diff']
    vmax_diff = kwargs['vmax_diff']
    nlevels_diff = kwargs['nlevels_diff']
    cmap_diff = get_cmap(kwargs['cmap_diff'])

    member = kwargs['member']
    k = kwargs['k']
    time = kwargs['time']

    state = State(c)
    rec_query = [id for id,r in state.info['fields'].items() if r['name']==vname and r['k']==k]
    assert len(rec_query)>0, f"cannot find state variable '{vname}' at k={k}"
    rec_id = rec_query[0]
    rec = state.info['fields'][rec_id]

    if c.debug:
        print(f"PID {c.pid:4} plotting state variable '{vname:20}' k={k:3} at {time} for member{member+1:03}", flush=True)

    ##if the viewer html file does not exist, generate it
    viewer = os.path.join(plot_dir, 'index.html')
    if not os.path.exists(viewer):
        generate_viewer_html(c, plot_dir, variables, figsize)

    ##plot the variables defined in kwargs, save to figfile
    figfile = os.path.join(plot_dir, f"{vname}_k{k}_{time:%Y%m%dT%H%M%S}_mem{member+1:03}.png")

    ##read the field from bin file in analysis dir
    var_prior = state.read_field(state.prior_file, c.grid.mask, member, rec_id)
    var_post = state.read_field(state.post_file, c.grid.mask, member, rec_id)
    incr = var_post - var_prior

    ##plot the field
    try:
        fig, ax = plt.subplots(1, 1, figsize=figsize)
        if rec['is_vector']:
            c.grid.plot_vectors(ax, incr, V=vmax_diff, showref=True, ref_units=rec['units'])
            adjust_ax_size(ax)

        else:
            c.grid.plot_field(ax, incr, vmin=vmin_diff, vmax=vmax_diff, cmap=cmap_diff)
            add_colorbar(fig, ax, cmap_diff, vmin_diff, vmax_diff, nlevels_diff, units=rec['units'])

        c.grid.plot_land(ax, color=landcolor)
        ax.set_title(f'analysis increment, member {member+1}', fontsize=16)
        ax.set_xlabel('x (m)', fontsize=14)
        ax.set_ylabel('y (m)', fontsize=14)
        fig.suptitle(f"{vname}, level {k:2}, {time}", fontsize=16)
        plt.savefig(figfile)
        plt.close()

    except Exception as e:
        print(f"ERROR: Failed to plot {vname} at level {k} and time {time} for member {member+1}")
        raise e

def generate_viewer_html(c, plot_dir, variables, figsize) -> None:
    """Generating a html page to help viewing the ensemble state variables"""
    if c.debug:
        print(f"Generating viewer.html page in {plot_dir}")

    with open(os.path.join(os.path.dirname(__file__), 'viewer.html'), 'rt') as f:
        html_page = f.read()

    state = State(c)

    ##replace the placeholder with the list of variables,levels,times,members
    levels_by_variable = ""
    times_by_variable = ""
    for vname in variables:
        levels_by_variable += f"'{vname}': ["
        levels = [r['k'] for id,r in state.info['fields'].items() if r['name']==vname]
        levels.sort()
        for level in levels:
            levels_by_variable += f"{level}, "
        levels_by_variable += "], \n"

        times_by_variable += f"'{vname}': ["
        for t in c.time + np.array(c.state_time_steps) * dt1h:
            times_by_variable += f"'{t:%Y%m%dT%H%M%S}', "
        times_by_variable += "], \n"

    html_page = html_page.replace("LEVELS_BY_VARIABLE", levels_by_variable)
    html_page = html_page.replace("TIMES_BY_VARIABLE", times_by_variable)

    members = "["
    for m in range(c.nens):
        members += f"'{m+1:03}', "
    members += "]"
    html_page = html_page.replace("MEMBERS", members)

    html_page = html_page.replace("TITLE", "Analysis Increments in Ensemble States")
    html_page = html_page.replace("IMAGE_WIDTH", f"{figsize[0]*60}")
    html_page = html_page.replace("IMAGE_HEIGHT", f"{figsize[1]*60}")

    ##write the html page to file
    with open(os.path.join(plot_dir, 'index.html'), 'w') as f:
        f.write(html_page)



import unittest
import os
import tempfile
from datetime import datetime, timezone
from NEDAS.config import Config

class TestConfig(unittest.TestCase):
    def test_convert_work_dir_to_abspath(self):
        c = Config()
        self.assertEqual(c.work_dir, os.path.abspath(c.work_dir))

    def test_nedas_root_available(self):
        c = Config()
        self.assertTrue(os.path.exists(c.nedas_root))

    def test_config_time_variable_type(self):
        c = Config()
        self.assertIsInstance(c.time, datetime)
        self.assertIsInstance(c.time_start, datetime)
        self.assertIsInstance(c.time_end, datetime)
        self.assertIsInstance(c.time_analysis_start, datetime)
        self.assertIsInstance(c.time_analysis_end, datetime)
        self.assertIsInstance(c.prev_time, datetime)
        self.assertIsInstance(c.next_time, datetime)

    def test_prev_next_time_arithmetic(self):
        tzinfo = timezone.utc
        time_start = datetime(2022, 1, 1, tzinfo=tzinfo)
        time = datetime(2023, 1, 1, tzinfo=tzinfo)
        cycle_period = 24
        c = Config(time_start=time_start, time=time, cycle_period=cycle_period)
        self.assertEqual(c.prev_time, datetime(2022, 12, 31, tzinfo=tzinfo))
        self.assertEqual(c.next_time, datetime(2023, 1, 2, tzinfo=tzinfo))

    def test_raise_exception_if_time_type_error(self):
        with self.assertRaises(TypeError):
            c = Config()
            c.time='2023-01-01'

    def test_argparse_time_str(self):
        tzinfo = timezone.utc
        c = Config(time='2001-01-01 00:00:00')
        self.assertEqual(c.time, datetime(2001,1,1, tzinfo=tzinfo))
        c = Config(time='20010101000000')
        self.assertEqual(c.time, datetime(2001,1,1, tzinfo=tzinfo))
        c = Config(time='2001-01-01T00:00:00Z')
        self.assertEqual(c.time, datetime(2001,1,1,tzinfo=tzinfo))

    def test_directory_names(self):
        dir_def = {'cycle_dir': '{work_dir}/cycle/{time:%Y%m%d}',
                   'analysis_dir': '{work_dir}/cycle/{time:%Y%m%d}/analysis',
                   'forecast_dir': '{work_dir}/cycle/{time:%Y%m%d}/{model_name}',}
        time = datetime(2023, 1, 1)
        cwd = os.getcwd()
        c = Config(work_dir='test', time=time, directories=dir_def)
        self.assertEqual(c.cycle_dir(time), os.path.join(cwd, 'test', 'cycle', '20230101'))
        self.assertEqual(c.analysis_dir(time), os.path.join(cwd, 'test', 'cycle', '20230101', 'analysis'))
        self.assertEqual(c.forecast_dir(time, 'qg'), os.path.join(cwd, 'test', 'cycle', '20230101', 'qg'))

    def test_nproc_mem_divides_nproc(self):
        with self.assertRaises(ValueError):
            c = Config(nproc=10, nproc_mem=3)

    def test_nproc_mem_nproc_rec_arithmetic(self):
        c = Config(nproc=10, nproc_mem=2)
        self.assertEqual(c.nproc_mem, 2)
        self.assertEqual(c.nproc_rec, 5)

    def test_yaml_file_dump_and_load(self):
        c = Config()

        with tempfile.NamedTemporaryFile(prefix='config', suffix='.yml', delete=False) as tmp_file:
            c.dump_yaml(tmp_file.name)
            c_tmp = Config(config_file=tmp_file.name)
            self.assertEqual(c.config_dict, c_tmp.config_dict)

if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod
import numpy as np

class Inflation(ABC):
    """
    Class for inflating the ensemble members (covariance inflation)
    """
    def __init__(self, coef: float=1.0,
                 adaptive: bool=False,
                 prior: bool=False, posterior: bool=False):
        self.coef = coef
        self.adaptive = adaptive
        self.prior = prior
        self.posterior = posterior

    def __call__(self, c, state, obs, flag):
        """
        Perform the covariance inflation method
        Input:
        -c: config object
        -state: State object with fields_prior, field_post
        -obs: Obs object with obs_seq, obs_prior_seq, obs_post_seq
        -flag: 'prior' or 'posterior'
        """
        if flag == 'prior' and self.prior:
            if self.adaptive:
                self.adaptive_prior_inflation(c, state, obs)
            self.apply_inflation(c, state, flag)

        if flag == 'posterior' and self.posterior:
            if self.adaptive:
                self.adaptive_post_inflation(c, state, obs)
            self.apply_inflation(c, state, flag)

    def obs_space_stats(self, c, state, obs):
        """observation-space statistics"""
        stats = {'total_nobs': 0,
                 'omb2': 0.0,  ##obs-minus-background differences squared
                 'omaamb': 0.0,
                 'amb2': 0.0,  ##analysis-minus-background diff squared
                 'varo': 0.0,  ##obs err variance
                 'varb': 0.0,  ##obs_prior (background) ensemble variances
                 'vara': 0.0,  ##obs_post (analysis) ensemble variances
                }

        ##go through each obs record
        for r, obs_rec_id in enumerate(obs.obs_rec_list[c.pid_rec]):
            obs_rec = obs.info['records'][obs_rec_id]
            nobs = obs_rec['nobs']

            ##1. get ensemble mean obs_prior:
            if obs_rec['is_vector']:
                nv = 2
                shape = (nv, nobs)
            else:
                nv = 1
                shape = (nobs,)

            ##sum over all obs_prior_seq locally stored on pid
            sum_obs_prior_pid = np.zeros(shape)
            for mem_id in state.mem_list[c.pid_mem]:
                sum_obs_prior_pid += obs.obs_prior_seq[mem_id, obs_rec_id]
            ##sum over all obs_prior_seq on differnet pids to get the total sum
            sum_obs_prior = c.comm_mem.allreduce(sum_obs_prior_pid)
            mean_obs_prior = sum_obs_prior / c.nens

            if obs.obs_post_seq:
                ##sum over all obs_prior_seq locally stored on pid
                sum_obs_post_pid = np.zeros(shape)
                for mem_id in state.mem_list[c.pid_mem]:
                    sum_obs_post_pid += obs.obs_post_seq[mem_id, obs_rec_id]
                ##sum over all obs_prior_seq on differnet pids to get the total sum
                sum_obs_post = c.comm_mem.allreduce(sum_obs_post_pid)
                mean_obs_post = sum_obs_post / c.nens

            ##2. get ensemble spread obs_prior:
            pert2_obs_prior_pid = np.zeros(shape)
            for mem_id in state.mem_list[c.pid_mem]:
                pert2_obs_prior_pid += (obs.obs_prior_seq[mem_id, obs_rec_id] - mean_obs_prior)**2
            pert2_obs_prior = c.comm_mem.allreduce(pert2_obs_prior_pid)
            variance_obs_prior = pert2_obs_prior / (c.nens - 1)

            if obs.obs_post_seq:
                pert2_obs_post_pid = np.zeros(shape)
                for mem_id in state.mem_list[c.pid_mem]:
                    pert2_obs_post_pid += (obs.obs_post_seq[mem_id, obs_rec_id] - mean_obs_post)**2
                pert2_obs_post = c.comm_mem.allreduce(pert2_obs_post_pid)
                variance_obs_post = pert2_obs_post / (c.nens - 1)

            obs_value = obs.obs_seq[obs_rec_id]['obs']
            stats['total_nobs'] += nv * nobs
            stats['omb2'] += np.sum((obs_value - mean_obs_prior)**2)
            stats['varo'] += np.sum(obs.obs_seq[obs_rec_id]['err_std']**2) * nv
            stats['varb'] += np.sum(variance_obs_prior)
            if obs.obs_post_seq:
                stats['amb2'] += np.sum((mean_obs_post - mean_obs_prior)**2)
                stats['omaamb'] += np.sum((obs_value - mean_obs_post)*(mean_obs_post - mean_obs_prior))
                stats['vara'] += np.sum(variance_obs_post)
        return stats

    @abstractmethod
    def adaptive_prior_inflation(self, c, state, obs):
        pass

    @abstractmethod
    def adaptive_post_inflation(self, c, state, obs):
        pass

    @abstractmethod
    def apply_inflation(self, c, state, flag):
        pass



##demo of how to run the qg model in standalone mode
from datetime import datetime
from NEDAS.models.qg import Model

def main():
    ##run the model by
    ##python models/qg/run.py -c models/qg/default.yml

    ##model config variables are listed in default.yml
    ##settings can also be changed in model_param
    model_param = {
        'kmax': 127,
        'psi_init_type': 'spectral_m',
        'initialize_energy': True,
        'model_env': '/cluster/home/yingyue/code/NEDAS/models/qg/env/setup_betzy.src',
        'model_code_dir': '/cluster/home/yingyue/code/NEDAS/models/qg',
        }

    run_opt = {
        'job_submit_cmd': '/cluster/home/yingyue/code/NEDAS/config/samples/job_submit_betzy.sh',
        'path': '/cluster/work/users/yingyue/qg_run',
        'time': datetime(2023,1,1),
        'forecast_period': 120,
        }

    model = Model(parse_args=True, **model_param)

    model.run(**run_opt)

if __name__ == '__main__':
    main()




import importlib

registry = {
  'ensemble': 'EnsembleCovariance',
  # 'static'
}

def get_covariance(c):

    if c.covariance not in registry.keys():
        raise NotImplementedError(f"Covariance model '{c.covariance}' not implemented")

    module = importlib.import_module("NEDAS.assim_tools.covariance."+c.covariance)

    Covariance = getattr(module, registry[c.covariance])

    return Covariance(c)



import numpy as np
from NEDAS.utils.njit import njit

@njit
def gradx(fld, dx, cyclic_dim=None):
    """Gradient of input field in x direction

    Args:
        fld (np.ndarray): input field, last two dimensions (ny, nx)
        dx (int): grid spacing in x, fld.shape
        cyclic_dim (str, optional): string 'x', 'y', 'xy', indicating the dimension(s) that are cyclic.

    Returns:
        np.ndarray: gradx of fld with same shape
    """
    fld_gradx = np.zeros(fld.shape)
    if cyclic_dim is not None and 'x' in cyclic_dim:
        ##centered difference of the neighboring grid points
        fld_gradx[..., :-1] = fld[..., 1:]  ##right neighbor
        fld_gradx[..., -1]  = fld[..., 0]
        fld_gradx[..., 1:] -= fld[..., :-1] ##left neighbor
        fld_gradx[..., 0]  -= fld[..., -1]
        fld_gradx /= 2.*dx
    else:
        ##centered difference for the middle part
        fld_gradx[..., 1:-1] = (fld[..., 2:] - fld[..., :-2]) / 2.0
        ##one-sided difference for the left,right edge points
        fld_gradx[..., 0] = (fld[..., 1] - fld[..., 0])
        fld_gradx[..., -1] = (fld[..., -1] - fld[..., -2])
        fld_gradx /= dx
    return fld_gradx

@njit
def grady(fld, dy, cyclic_dim=None):
    """gradient of input fld in y direction, similar to gradx"""
    fld_grady = np.zeros(fld.shape)
    if cyclic_dim is not None and 'y' in cyclic_dim:
        ##centered difference of the neighboring grid points
        fld_grady[..., :-1, :] = fld[..., 1:, :]
        fld_grady[..., -1, :]  = fld[..., 0, :]
        fld_grady[..., 1:, :] -= fld[..., :-1, :]
        fld_grady[..., 0, :]  -= fld[..., -1, :]
        fld_grady /= 2.*dy
    else:
        ##centered difference for the middle part
        fld_grady[..., 1:-1, :] = (fld[..., 2:, :] - fld[..., :-2, :]) / 2.0
        ##one-sided difference for the left,right edge points
        fld_grady[..., 0, :] = (fld[..., 1, :] - fld[..., 0, :])
        fld_grady[..., -1, :] = (fld[..., -1, :] - fld[..., -2, :])
        fld_grady /= dy
    return fld_grady

@njit
def gradx2(fld, dx, cyclic_dim=None):
    return gradx(gradx(fld, dx, cyclic_dim), dx, cyclic_dim)

@njit
def grady2(fld, dy, cyclic_dim=None):
    return grady(grady(fld, dy, cyclic_dim), dy, cyclic_dim)

@njit
def gradxy(fld, dx, dy, cyclic_dim=None):
    return grady(gradx(fld, dx, cyclic_dim), dy, cyclic_dim)

@njit
def laplacian(fld, dx, dy, cyclic_dim=None):
    return gradx2(fld, dx, cyclic_dim) + grady2(fld, dy, cyclic_dim)


def coarsen(grid, fld, nlevel):
    """
    Coarsen the image by downsampling the grid points by factors of 1/2,

    Args:
        grid (Grid): the original grid
        fld (np.ndarray): field with last two dimensions ny,nx
        nlevel (int): number of resolution levels to go down

    Returns:
        Grid: new grid with lower resolution
        np.ndarray: new field defined on the new grid
    """
    assert grid.x.shape == fld.shape[-2:], "coarsen: input fld size mismatch with grid"
    if nlevel == 0:
        return grid, fld
    grid1 = grid.change_resolution_level(nlevel)
    grid.set_destination_grid(grid1)
    fld1 = np.zeros(fld.shape[:-2]+(grid1.ny, grid1.nx))
    for ind in np.ndindex(fld.shape[:-2]):
        fld1[ind] = grid.convert(fld[ind], is_vector=False, method='linear', coarse_grain=True)
    return grid1, fld1


def refine(grid, mask, fld, nlevel):
    """
    Refine the image by upsampling the grid points by factors of 2,

    Args:
        grid (Grid): the original grid
        fld (np.ndarray): field with last two dimensions ny,nx
        nlevel (int): number of resolution levels to go up

    Returns:
        Grid: new grid with higher resolution
        np.ndarray: new field on new grid
    """
    assert grid.x.shape == fld.shape[-2:], "refine: input fld size mismatch with grid"
    if nlevel == 0:
        return grid, fld
    grid1 = grid.change_resolution_level(-nlevel)
    grid1.mask = mask  ##don't know how to sharpen mask, need to provide high-res mask from input args
    grid.set_destination_grid(grid1)
    fld1 = np.zeros(fld.shape[:-2]+(grid1.ny, grid1.nx))
    for ind in np.ndindex(fld.shape[:-2]):
        fld_nearest = grid.convert(fld[ind], is_vector=False, method='nearest')
        fld_linear = grid.convert(fld[ind], is_vector=False, method='linear')
        void = np.isnan(fld_linear)
        fld_linear[void] = fld_nearest[void]
        fld1[ind] = fld_linear
    return grid1, fld1

##some original coarsen/sharpen functions working for 2**n grid points
def coarsen_mask(mask, lev1, lev2):  ##only subsample no smoothing, avoid mask growing
    if lev1 < lev2:
        for k in range(lev1, lev2):
            ni, nj = mask.shape[-2:]
            mask1 = mask[..., 0:ni:2, 0:nj:2]
            mask = mask1
    return mask

def coarsen_field(fld, lev1, lev2):
    if lev1 < lev2:
        for k in range(lev1, lev2):
            ni, nj = fld.shape[-2:]
            fld1 = 0.25*(fld[..., 0:ni:2, 0:nj:2] + fld[..., 1:ni:2, 0:nj:2] + fld[..., 0:ni:2, 1:nj:2] + fld[..., 1:ni:2, 1:nj:2])
            fld = fld1
    return fld

def sharpen_field(fld, lev1, lev2):
    if lev1 > lev2:
        for k in range(lev1, lev2, -1):
            ni, nj = fld.shape[-2:]
            fld1 = np.zeros(fld.shape[:-2]+(ni*2, nj))
            fld1[..., 0:ni*2:2, :] = fld
            fld1[..., 1:ni*2:2, :] = 0.5*(np.roll(fld, -1, axis=-2) + fld)
            fld2 = np.zeros(fld.shape[:-2]+(ni*2, nj*2))
            fld2[..., :, 0:nj*2:2] = fld1
            fld2[..., :, 1:nj*2:2] = 0.5*(np.roll(fld1, -1, axis=-1) + fld1)
            fld = fld2
    return fld

def warp(grid, fld, u, v):
    """
    Warp the image with input vector field

    Args:
        grid (Grid): the grid on which the image is defined
        fld (np.ndarray): input image
        u (np.ndarray): displacement vector x component, in :code:`grid.x` units
        v (np.ndarray): displacement vector y component, in :code:`grid.y` units

    Returns:
        np.ndarray: the warped image
    """
    assert grid.x.shape == fld.shape[-2:], "warp: input fld size mismatch with grid"
    assert grid.x.shape == u.shape, "warp: input u size mismatch with grid"
    assert grid.x.shape == v.shape, "warp: input v size mismatch with grid"
    fld1 = fld.copy()
    for ind in np.ndindex(fld.shape[:-2]):
        fld_nearest = grid.interp(fld[ind], grid.x-u, grid.y-v, method='nearest')
        fld_linear = grid.interp(fld[ind], grid.x-u, grid.y-v, method='linear')
        void = np.isnan(fld_linear)
        fld_linear[void] = fld_nearest[void]
        fld1[ind] = fld_linear
    return fld1

